from typing import Any, Dict, Optional, TYPE_CHECKING, Tuple, Union

from _moderngl import InvalidObject

from .buffer import Buffer
from .renderbuffer import Renderbuffer
from .texture import Texture

if TYPE_CHECKING:
    from .context import Context

__all__ = ['Framebuffer']


class Framebuffer:
    """
    A :py:class:`Framebuffer` is a collection of buffers that can be used as the destination for rendering.

    The buffers for Framebuffer objects reference images from either Textures or Renderbuffers.
    Create a :py:class:`Framebuffer` using :py:meth:`Context.framebuffer`.
    """

    __slots__ = [
        'mglo', '_color_attachments', '_depth_attachment', '_size', '_samples', '_glo',
        'ctx', '_is_reference', 'extra'
    ]

    def __init__(self):
        self.mglo = None  #: Internal representation for debug purposes only.
        self._color_attachments = None
        self._depth_attachment = None
        self._size = (None, None)
        self._samples: int = None
        self._glo: int = None
        self.ctx: Context = None  #: The context this object belongs to
        self._is_reference = None  #: Detected framebuffers we should not delete
        self.extra: Any = None  #: Attribute for storing user defined objects
        raise TypeError()

    def __repr__(self) -> str:
        if hasattr(self, '_glo'):
            return '<Framebuffer: %d>' % self._glo
        else:
            return "<Framebuffer: INCOMPLETE>"

    def __eq__(self, other: Any) -> bool:
        return type(self) is type(other) and self.mglo is other.mglo

    def __hash__(self) -> int:
        return id(self)

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        # Don't delete default framebuffer or a reference
        if self._is_reference:
            return

        # If object was initialized properly (ctx present) and gc_mode is auto
        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def viewport(self) -> Tuple[int, int, int, int]:
        """tuple: Get or set the viewport of the framebuffer."""
        return self.mglo.viewport

    @viewport.setter
    def viewport(self, value: Tuple[int, int, int, int]) -> None:
        x, y, w, h = value
        self.mglo.viewport = (int(x), int(y), int(w), int(h))

    @property
    def scissor(self) -> Tuple[int, int, int, int]:
        """
        tuple: Get or set the scissor box of the framebuffer.

        When scissor testing is enabled fragments outside
        the defined scissor box will be discarded. This
        applies to rendered geometry or :py:meth:`Framebuffer.clear`.

        Setting is value enables scissor testing in the framebuffer.
        Setting the scissor to ``None`` disables scissor testing
        and reverts the scissor box to match the framebuffer size.

        Example::

            # Enable scissor testing
            >>> ctx.scissor = 100, 100, 200, 100
            # Disable scissor testing
            >>> ctx.scissor = None
        """
        return self.mglo.scissor

    @scissor.setter
    def scissor(self, value: Tuple[int, int, int, int]) -> None:
        if value is None:
            self.mglo.scissor = None
        else:
            x, y, w, h = value
            self.mglo.scissor = (int(x), int(y), int(w), int(h))

    @property
    def color_mask(self) -> Tuple[bool, bool, bool, bool]:
        """
        tuple: The color mask of the framebuffer.

        Color masking controls what components in color attachments will be
        affected by fragment write operations.
        This includes rendering geometry and clearing the framebuffer.

        Default value: ``(True, True, True, True)``.

        Examples::

            # Block writing to all color components (rgba) in color attachments
            fbo.color_mask = False, False, False, False

            # Re-enable writing to color attachments
            fbo.color_mask = True, True, True, True

            # Block fragment writes to alpha channel
            fbo.color_mask = True, True, True, False
        """
        return self.mglo.color_mask

    @color_mask.setter
    def color_mask(self, value: Tuple[bool, bool, bool, bool]) -> None:
        self.mglo.color_mask = value

    @property
    def depth_mask(self) -> bool:
        """
        bool: The depth mask of the framebuffer.

        Depth mask enables or disables write operations to the depth buffer.
        This also applies when clearing the framebuffer.
        If depth testing is enabled fragments will still be culled, but
        the depth buffer will not be updated with new values. This is
        a very useful tool in many rendering techniques.

        Default value: ``True``
        """
        return self.mglo.depth_mask

    @depth_mask.setter
    def depth_mask(self, value: bool) -> None:
        self.mglo.depth_mask = value

    @property
    def width(self) -> int:
        """
        int: The width of the framebuffer.

        Framebuffers created by a window will only report its initial size.
        It's better get size information from the window itself.
        """
        return self._size[0]

    @property
    def height(self) -> int:
        """
        int: The height of the framebuffer.

        Framebuffers created by a window will only report its initial size.
        It's better get size information from the window itself.
        """
        return self._size[1]

    @property
    def size(self) -> tuple:
        """
        tuple: The size of the framebuffer.

        Framebuffers created by a window will only report its initial size.
        It's better get size information from the window itself.
        """
        return self._size

    @property
    def samples(self) -> int:
        """int: The samples of the framebuffer."""
        return self._samples

    @property
    def bits(self) -> Dict[str, str]:
        """dict: The bits of the framebuffer."""
        return self.mglo.bits

    @property
    def color_attachments(self) -> Tuple[Union[Texture, Renderbuffer], ...]:
        """tuple: The color attachments of the framebuffer."""
        return self._color_attachments

    @property
    def depth_attachment(self) -> Union[Texture, Renderbuffer]:
        """Texture or Renderbuffer: The depth attachment of the framebuffer."""
        return self._depth_attachment

    @property
    def glo(self) -> int:
        """
        int: The internal OpenGL object.

        This values is provided for debug purposes only.
        """
        return self._glo

    def clear(
        self,
        red: float = 0.0,
        green: float = 0.0,
        blue: float = 0.0,
        alpha: float = 0.0,
        depth: float = 1.0,
        *,
        viewport: Optional[Union[Tuple[int, int], Tuple[int, int, int, int]]] = None,
        color: Optional[Tuple[float, float, float, float]] = None,
    ) -> None:
        """
        Clear the framebuffer.

        If a `viewport` passed in, a scissor test will be used to clear the given viewport.
        This viewport take presence over the framebuffers :py:attr:`~moderngl.Framebuffer.scissor`.
        Clearing can still be done with scissor if no viewport is passed in.

        This method also respects the
        :py:attr:`~moderngl.Framebuffer.color_mask` and
        :py:attr:`~moderngl.Framebuffer.depth_mask`. It can for example be used to only clear
        the depth or color buffer or specific components in the color buffer.

        If the `viewport` is a 2-tuple it will clear the
        ``(0, 0, width, height)`` where ``(width, height)`` is the 2-tuple.

        If the `viewport` is a 4-tuple it will clear the given viewport.

        Args:
            red (float): color component.
            green (float): color component.
            blue (float): color component.
            alpha (float): alpha component.
            depth (float): depth value.

        Keyword Args:
            viewport (tuple): The viewport.
            color (tuple): Optional tuple replacing the red, green, blue and alpha arguments
        """
        if color is not None:
            red, green, blue, alpha, *_ = tuple(color) + (0.0, 0.0, 0.0, 0.0)

        if viewport is not None:
            viewport = tuple(viewport)  # type: ignore

        self.mglo.clear(red, green, blue, alpha, depth, viewport)

    def use(self) -> None:
        """Bind the framebuffer. Sets the target for rendering commands."""
        self.ctx.fbo = self
        self.mglo.use()

    def read(
        self,
        viewport: Optional[Union[Tuple[int, int], Tuple[int, int, int, int]]] = None,
        components: int = 3,
        *,
        attachment: int = 0,
        alignment: int = 1,
        dtype: str = 'f1',
        clamp: bool = False,
    ) -> bytes:
        """
        Read the content of the framebuffer.

        .. code:: python

            # Read the first color attachment's RGBA data
            data = fbo.read(components=4)
            # Read the second color attachment's RGB data
            data = fbo.read(attachment=1)
            # Read the depth attachment
            data = fbo.read(attachment=-1)
            # Read the lower left 10 x 10 pixels from the first color attachment
            data = fbo.read(viewport=(0, 0, 10, 10))

        Args:
            viewport (tuple): The viewport.
            components (int): The number of components to read.

        Keyword Args:
            attachment (int): The color attachment number. -1 for the depth attachment
            alignment (int): The byte alignment of the pixels.
            dtype (str): Data type.
            clamp (bool): Clamps floating point values to ``[0.0, 1.0]``

        Returns:
            bytes
        """
        return self.mglo.read(viewport, components, attachment, alignment, clamp, dtype)

    def read_into(
        self,
        buffer: Any,
        viewport: Optional[Union[Tuple[int, int], Tuple[int, int, int, int]]] = None,
        components: int = 3,
        *,
        attachment: int = 0,
        alignment: int = 1,
        dtype: str = 'f1',
        write_offset: int = 0,
    ) -> None:
        """
        Read the content of the framebuffer into a buffer.

        Args:
            buffer (bytearray): The buffer that will receive the pixels.
            viewport (tuple): The viewport.
            components (int): The number of components to read.

        Keyword Args:
            attachment (int): The color attachment.
            alignment (int): The byte alignment of the pixels.
            dtype (str): Data type.
            write_offset (int): The write offset.
        """
        if type(buffer) is Buffer:
            buffer = buffer.mglo

        return self.mglo.read_into(buffer, viewport, components, attachment, alignment, dtype, write_offset)

    def release(self) -> None:
        """Release the ModernGL object."""
        if self.mglo is not None:
            self._color_attachments = None
            self._depth_attachment = None
            self.mglo.release()
            self.mglo = InvalidObject()



'''
    Renders a blue triangle
'''

import numpy as np

import moderngl
from ported._example import Example


class PerspectiveProjection(Example):
    gl_version = (3, 3)
    title = "Perspective Projection"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec3 vert;

                uniform float z_near;
                uniform float z_far;
                uniform float fovy;
                uniform float ratio;

                uniform vec3 center;
                uniform vec3 eye;
                uniform vec3 up;

                mat4 perspective() {
                    float zmul = (-2.0 * z_near * z_far) / (z_far - z_near);
                    float ymul = 1.0 / tan(fovy * 3.14159265 / 360);
                    float xmul = ymul / ratio;

                    return mat4(
                        xmul, 0.0, 0.0, 0.0,
                        0.0, ymul, 0.0, 0.0,
                        0.0, 0.0, -1.0, -1.0,
                        0.0, 0.0, zmul, 0.0
                    );
                }

                mat4 lookat() {
                    vec3 forward = normalize(center - eye);
                    vec3 side = normalize(cross(forward, up));
                    vec3 upward = cross(side, forward);

                    return mat4(
                        side.x, upward.x, -forward.x, 0,
                        side.y, upward.y, -forward.y, 0,
                        side.z, upward.z, -forward.z, 0,
                        -dot(eye, side), -dot(eye, upward), dot(eye, forward), 1
                    );
                }

                void main() {
                    gl_Position = perspective() * lookat() * vec4(vert, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 color;

                void main() {
                    color = vec4(0.04, 0.04, 0.04, 1.0);

                }
            ''',
        )

        self.prog['z_near'].value = 0.1
        self.prog['z_far'].value = 1000.0
        self.prog['ratio'].value = self.aspect_ratio
        self.prog['fovy'].value = 60

        self.prog['eye'].value = (3, 3, 3)
        self.prog['center'].value = (0, 0, 0)
        self.prog['up'].value = (0, 0, 1)

        grid = []

        for i in range(65):
            grid.append([i - 32, -32.0, 0.0, i - 32, 32.0, 0.0])
            grid.append([-32.0, i - 32, 0.0, 32.0, i - 32, 0.0])

        grid = np.array(grid, dtype='f4')

        self.vbo = self.ctx.buffer(grid)
        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, 'vert')

    def render(self, time: float, frame_time: float):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.vao.render(moderngl.LINES, 65 * 4)


if __name__ == '__main__':
    PerspectiveProjection.run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import ModernGL
# from PIL import Image

# render_size = 1024, 1024
# image_size = 512, 512

# ctx = ModernGL.create_standalone_context()

# color_rbo = ctx.renderbuffer(render_size)
# depth_rbo = ctx.depth_renderbuffer(render_size)
# fbo = ctx.framebuffer(color_rbo, depth_rbo)

# fbo.use()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         uniform vec2 WindowSize;

#         in vec2 in_vert;
#         in vec3 in_color;

#         out vec3 v_color;

#         void main() {
#             v_color = in_color;
#             gl_Position = vec4(in_vert / WindowSize * 2.0, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         in vec3 v_color;
#         out vec4 f_color;

#         void main() {
#             f_color = vec4(v_color, 1.0);
#         }
#     '''),
# ])

# window_size = prog.uniforms['WindowSize']

# vbo = ctx.buffer(struct.pack(
#     '15f',
#     0.0, 100.0, 1.0, 0.0, 0.0,
#     -86.0, -50.0, 0.0, 1.0, 0.0,
#     86.0, -50.0, 0.0, 0.0, 1.0,
# ))

# vao = ctx.simple_vertex_array(prog, vbo, ['in_vert', 'in_color'])

# width, height = render_size
# ctx.viewport = (0, 0, width, height)
# ctx.clear(0.9, 0.9, 0.9)
# ctx.enable(ModernGL.BLEND)
# window_size.value = (width, height)
# vao.render()

# img = Image.frombytes('RGB', render_size, fbo.read())
# img = img.transpose(Image.FLIP_TOP_BOTTOM)
# img = img.resize(image_size, Image.LANCZOS)
# img.save('window_coordinates.png')




'''
Example of using shader storage buffer in compute shader.
We read from a buffer and write the result to another buffer.
Every frame we swap the buffers around transforming positions
of balls.

Buffer.bind_to_storage_buffer is used to bind a buffer as storage buffer
to a specific binding point specified in the compute program.

In addition we render the balls using a geometry shader to easily
batch draw them all in one render call.

author: minu jeong
modified by: einarf
'''
import math
import random
import numpy as np
from ported._example import Example


items_vertex_shader_code = """
    #version 430

    in vec4 in_vert;
    in vec4 in_col;

    out vec4 v_color;

    void main()
    {
        gl_Position = in_vert; // x, y, 0, radius
        v_color = in_col;
    }
    """

# Geometry shader turning the points into triangle strips.
# This can also be done with point sprites.
items_geo_shader = """
    #version 330

    layout(points) in;
    layout(triangle_strip, max_vertices=4) out;

    in vec4 v_color[];
    out vec2 uv;
    out vec4 color;

    void main() {
        float radius = gl_in[0].gl_Position.w;
        vec2 pos = gl_in[0].gl_Position.xy;

        // Emit the triangle strip creating a "quad"
        // Lower left
        gl_Position = vec4(pos + vec2(-radius, -radius), 0, 1);
        color = v_color[0];
        uv = vec2(0, 0);
        EmitVertex();

        // upper left
        gl_Position = vec4(pos + vec2(-radius, radius), 0, 1);
        color = v_color[0];
        uv = vec2(0, 1);
        EmitVertex();

        // lower right
        gl_Position = vec4(pos + vec2(radius, -radius), 0, 1);
        color = v_color[0];
        uv = vec2(1, 0);
        EmitVertex();

        // upper right
        gl_Position = vec4(pos + vec2(radius, radius), 0, 1);
        color = v_color[0];
        uv = vec2(1, 1);
        EmitVertex();

        EndPrimitive();
    }
"""

items_fragment_shader_code = """
    #version 430

    in vec2 uv;
    in vec4 color;
    out vec4 out_color;
    void main()
    {
        // Calculate the length from the center of the "quad"
        // using texture coordinates discarding fragments
        // further away than 0.5 creating a circle.
        if (length(vec2(0.5, 0.5) - uv.xy) > 0.5)
        {
            discard;
        }
        out_color = color;
    }
"""

# calc position with compute shader
compute_worker_shader_code = """
#version 430
#define GROUP_SIZE %COMPUTE_SIZE%

layout(local_size_x=GROUP_SIZE) in;

// All values are vec4s because of block alignment rules (keep it simple).
// We could also declare all values as floats to make it tightly packed.
// See : https://www.khronos.org/opengl/wiki/Interface_Block_(GLSL)#Memory_layout
struct Ball
{
    vec4 pos; // x, y, 0, radius
    vec4 vel; // x, y (velocity)
    vec4 col; // r, g, b (color)
};

layout(std430, binding=0) buffer balls_in
{
    Ball balls[];
} In;
layout(std430, binding=1) buffer balls_out
{
    Ball balls[];
} Out;

void main()
{
    int x = int(gl_GlobalInvocationID);

    Ball in_ball = In.balls[x];

    vec4 p = in_ball.pos.xyzw;
    vec4 v = in_ball.vel.xyzw;

    p.xy += v.xy;

    float rad = p.w * 0.5;
    if (p.x - rad <= -1.0)
    {
        p.x = -1.0 + rad;
        v.x *= -0.98;
    }
    else if (p.x + rad >= 1.0)
    {
        p.x = 1.0 - rad;
        v.x *= -0.98;
    }

    if (p.y - rad <= -1.0)
    {
        p.y = -1.0 + rad;
        v.y *= -0.98;
    }
    else if (p.y + rad >= 1.0)
    {
        p.y = 1.0 - rad;
        v.y *= -0.98;
    }
    v.y += -0.001;

    Ball out_ball;
    out_ball.pos.xyzw = p.xyzw;
    out_ball.vel.xyzw = v.xyzw;

    vec4 c = in_ball.col.xyzw;
    out_ball.col.xyzw = c.xyzw;

    Out.balls[x] = out_ball;
}
"""

class ComputeShaderSSBO(Example):
    title = "Compute Shader SSBO"
    gl_version = 4, 3  # Required opengl version
    window_size = 600, 600  # Initial window size
    aspect_ratio = 1.0  # Force viewport aspect ratio (regardless of window size)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.COUNT = 256  # number of balls
        self.STRUCT_SIZE = 12  # number of floats per item/ball

        # Program for drawing the balls / items
        self.program = self.ctx.program(
            vertex_shader=items_vertex_shader_code,
            geometry_shader=items_geo_shader,
            fragment_shader=items_fragment_shader_code
        )

        # Load compute shader
        compute_shader_code_parsed = compute_worker_shader_code.replace("%COMPUTE_SIZE%", str(self.COUNT))
        self.compute_shader = self.ctx.compute_shader(compute_shader_code_parsed)

        # Create the two buffers the compute shader will write and read from
        compute_data = np.fromiter(self.gen_initial_data(), dtype="f4")
        self.compute_buffer_a = self.ctx.buffer(compute_data)
        self.compute_buffer_b = self.ctx.buffer(compute_data)

        # Prepare vertex arrays to drawing balls using the compute shader buffers are input
        # We use 4x4 (padding format) to skip the velocity data (not needed for drawing the balls)
        self.balls_a = self.ctx.vertex_array(
            self.program, [(self.compute_buffer_a, '4f 4x4 4f', 'in_vert', 'in_col')],
        )
        self.balls_b = self.ctx.vertex_array(
            self.program, [(self.compute_buffer_b, '4f 4x4 4f', 'in_vert', 'in_col')],
        )

    def gen_initial_data(self):
        """Generator function creating the initial buffer data"""
        for i in range(self.COUNT):
            _angle = (i / self.COUNT) * math.pi * 2.0
            _dist = 0.125
            radius = random.random() * 0.01 + 0.01
            # position and radius (vec4)
            yield math.cos(_angle) * _dist
            yield math.sin(_angle) * _dist
            yield 0.0
            yield radius
            # velocity (vec4)
            _v = random.random() * 0.005 + 0.01
            yield math.cos(_angle) * _v
            yield math.sin(_angle) * _v
            yield 0.0
            yield 0.0
            # color (vec4)
            yield 1.0 * random.random()
            yield 1.0 * random.random()
            yield 1.0 * random.random()
            yield 1.0


    def render(self, time, frame_time):
        # Calculate the next position of the balls with compute shader
        self.compute_buffer_a.bind_to_storage_buffer(0)
        self.compute_buffer_b.bind_to_storage_buffer(1)
        self.compute_shader.run(group_x=self.STRUCT_SIZE)

        # Batch draw the balls
        self.balls_b.render(mode=self.ctx.POINTS)

        # Swap the buffers and vertex arrays around for next frame
        self.compute_buffer_a, self.compute_buffer_b = self.compute_buffer_b, self.compute_buffer_a
        self.balls_a, self.balls_b = self.balls_b, self.balls_a


if __name__ == "__main__":
    ComputeShaderSSBO.run()



import struct

import ModernGL
from PyQt5 import QtOpenGL, QtWidgets


class QGLControllerWidget(QtOpenGL.QGLWidget):
    def __init__(self):
        fmt = QtOpenGL.QGLFormat()
        fmt.setVersion(3, 3)
        fmt.setProfile(QtOpenGL.QGLFormat.CoreProfile)
        fmt.setSampleBuffers(True)
        super(QGLControllerWidget, self).__init__(fmt, None)

    def initializeGL(self):
        self.ctx = ModernGL.create_context()

        prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 vert;

                void main() {
                    gl_Position = vec4(vert, 0.0, 1.0);
                }
            '''),
            fragment_shader='''
                #version 330

                out vec4 color;

                void main() {
                    color = vec4(0.30, 0.50, 1.00, 1.0);
                }
            '''),
        ])

        vbo = self.ctx.buffer(struct.pack('6f', 0.0, 0.8, -0.6, -0.8, 0.6, -0.8))
        self.vao = self.ctx.simple_vertex_array(prog, vbo, ['vert'])

    def paintGL(self):
        self.ctx.viewport = (0, 0, self.width(), self.height())
        self.ctx.clear(0.9, 0.9, 0.9)
        self.vao.render()
        self.ctx.finish()


app = QtWidgets.QApplication([])
window = QGLControllerWidget()
window.move(QtWidgets.QDesktopWidget().rect().center() - window.rect().center())
window.show()
app.exec_()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import ModernGL
# from PIL import Image

# render_size = 1024, 1024
# image_size = 512, 512

# ctx = ModernGL.create_standalone_context()

# color_rbo = ctx.renderbuffer(render_size)
# depth_rbo = ctx.depth_renderbuffer(render_size)
# fbo = ctx.framebuffer(color_rbo, depth_rbo)

# fbo.use()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 vert;
#         out vec2 tex;

#         void main() {
#             gl_Position = vec4(vert, 0.0, 1.0);
#             tex = vert / 2.0;
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         in vec2 tex;
#         out vec4 color;

#         void main() {
#             vec2 z = tex;

#             int i;
#             for(i = 0; i < 100; i++) {
#                 vec2 v = vec2((z.x * z.x - z.y * z.y), (z.y * z.x + z.x * z.y)) - vec2(0.64, -0.47);
#                 if (dot(v, v) > 4.0) break;
#                 z = v;
#             }

#             float cm = fract((i == 100 ? 0.0 : float(i)) * 10.0 / 100);
#             color = vec4(fract(cm + 0.0 / 3.0), fract(cm + 1.0 / 3.0), fract(cm + 2.0 / 3.0), 1.0);
#         }
#     ''')
# ])

# vbo = ctx.buffer(struct.pack('8f', -1.0, -1.0, -1.0, 1.0, 1.0, -1.0, 1.0, 1.0))
# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])

# vao.render(ModernGL.TRIANGLE_STRIP)

# img = Image.frombytes('RGB', render_size, fbo.read())
# img = img.transpose(Image.FLIP_TOP_BOTTOM)
# img = img.resize(image_size, Image.LANCZOS)
# img.save('Fractal.png')



import unittest

import moderngl
import numpy as np

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_1(self):
        prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec4 in_vert;
                out vec4 out_vert;

                void main() {
                    out_vert = in_vert;
                }
            ''',
            varyings=['out_vert']
        )

        vertices = [
            4.0, 2.0, 7.5, 1.8,
            3.0, 2.8, 6.0, 10.0
        ]
        count = 10
        indices = [0, 1] * 10

        # UNSIGNED_INT index
        vbo1 = self.ctx.buffer(np.array(vertices, dtype='f4').tobytes())
        vbo2 = self.ctx.buffer(reserve=vbo1.size * count)
        index = self.ctx.buffer(np.array(indices, dtype='u4').tobytes())
        vao = self.ctx.simple_vertex_array(prog, vbo1, 'in_vert', index_buffer=index, index_element_size=4)
        vao.transform(vbo2, moderngl.POINTS)
        res = np.frombuffer(vbo2.read(), dtype='f4')
        np.testing.assert_almost_equal(res, vertices * count)

        # UNSIGNED_SHORT index
        vbo1 = self.ctx.buffer(np.array(vertices, dtype='f4').tobytes())
        vbo2 = self.ctx.buffer(reserve=vbo1.size * count)
        index = self.ctx.buffer(np.array(indices, dtype='u2').tobytes())
        vao = self.ctx.simple_vertex_array(prog, vbo1, 'in_vert', index_buffer=index, index_element_size=2)
        vao.transform(vbo2, moderngl.POINTS)
        res = np.frombuffer(vbo2.read(), dtype='f4')
        np.testing.assert_almost_equal(res, vertices * count)

        # UNSIGNED_BYTE index
        vbo1 = self.ctx.buffer(np.array(vertices, dtype='f4').tobytes())
        vbo2 = self.ctx.buffer(reserve=vbo1.size * count)
        index = self.ctx.buffer(np.array(indices, dtype='u1').tobytes())
        vao = self.ctx.simple_vertex_array(prog, vbo1, 'in_vert', index_buffer=index, index_element_size=1)
        vao.transform(vbo2, moderngl.POINTS)
        res = np.frombuffer(vbo2.read(), dtype='f4')
        np.testing.assert_almost_equal(res, vertices * count)

    def test_2(self):
        prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec4 in_vert;
                out vec4 out_vert;

                void main() {
                    out_vert = in_vert;
                }
            ''',
            varyings=['out_vert']
        )

        vertices = [
            4.0, 2.0, 7.5, 1.8,
            3.0, 2.8, 6.0, 10.0
        ]
        indices = [0, 1, 0, 1, 0, 1, 0, 1, 0]

        vbo1 = self.ctx.buffer(np.array(vertices, dtype='f4').tobytes())
        index_u4 = self.ctx.buffer(np.array(indices, dtype='u4').tobytes())

        with self.assertRaises(moderngl.Error):
            self.ctx.simple_vertex_array(prog, vbo1, 'in_vert', index_buffer=index_u4, index_element_size=0)


if __name__ == '__main__':
    unittest.main()



from pyrr import Matrix44

import moderngl
from ported._example import Example


class LoadingOBJ(Example):
    title = "Loading OBJ"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.obj = self.load_scene('sitting_dummy.obj')
        self.texture = self.load_texture_2d('wood.jpg')

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 Mvp;

                in vec3 in_position;
                in vec3 in_normal;
                in vec2 in_texcoord_0;

                out vec3 v_vert;
                out vec3 v_norm;
                out vec2 v_text;

                void main() {
                    v_vert = in_position;
                    v_norm = in_normal;
                    v_text = in_texcoord_0;
                    gl_Position = Mvp * vec4(in_position, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                uniform sampler2D Texture;
                uniform vec4 Color;
                uniform vec3 Light;

                in vec3 v_vert;
                in vec3 v_norm;
                in vec2 v_text;

                out vec4 f_color;

                void main() {
                    float lum = dot(normalize(v_norm), normalize(v_vert - Light));
                    lum = acos(lum) / 3.14159265;
                    lum = clamp(lum, 0.0, 1.0);
                    lum = lum * lum;
                    lum = smoothstep(0.0, 1.0, lum);
                    lum *= smoothstep(0.0, 80.0, v_vert.z) * 0.3 + 0.7;
                    lum = lum * 0.8 + 0.2;

                    vec3 color = texture(Texture, v_text).rgb;
                    color = color * (1.0 - Color.a) + Color.rgb * Color.a;
                    f_color = vec4(color * lum, 1.0);
                }
            ''',
        )

        self.light = self.prog['Light']
        self.color = self.prog['Color']
        self.mvp = self.prog['Mvp']

        # Create a vao from the first root node (attribs are auto mapped)
        self.vao = self.obj.root_nodes[0].mesh.vao.instance(self.prog)

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.DEPTH_TEST)

        proj = Matrix44.perspective_projection(45.0, self.aspect_ratio, 0.1, 1000.0)
        lookat = Matrix44.look_at(
            (-85, -180, 140),
            (0.0, 0.0, 65.0),
            (0.0, 0.0, 1.0),
        )

        self.light.value = (-140.0, -300.0, 350.0)
        self.color.value = (1.0, 1.0, 1.0, 0.25)
        self.mvp.write((proj * lookat).astype('f4'))

        self.texture.use()
        self.vao.render()


if __name__ == '__main__':
    LoadingOBJ.run()



import numpy as np
from pyrr import Matrix44

import moderngl
from ported._example import Example


def grid(size, steps):
    u = np.repeat(np.linspace(-size, size, steps), 2)
    v = np.tile([-size, size], steps)
    w = np.zeros(steps * 2)
    return np.concatenate([np.dstack([u, v, w]), np.dstack([v, u, w])])


class SimpleGrid(Example):
    title = "Simple Grid"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 Mvp;

                in vec3 in_vert;

                void main() {
                    gl_Position = Mvp * vec4(in_vert, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 f_color;

                void main() {
                    f_color = vec4(0.1, 0.1, 0.1, 1.0);
                }
            ''',
        )

        self.mvp = self.prog['Mvp']

        self.vbo = self.ctx.buffer(grid(15, 10).astype('f4'))
        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, 'in_vert')

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.DEPTH_TEST)

        proj = Matrix44.perspective_projection(45.0, self.aspect_ratio, 0.1, 1000.0)
        lookat = Matrix44.look_at(
            (40.0, 30.0, 30.0),
            (0.0, 0.0, 0.0),
            (0.0, 0.0, 1.0),
        )

        self.mvp.write((proj * lookat).astype('f4'))
        self.vao.render(moderngl.LINES)


if __name__ == '__main__':
    SimpleGrid.run()



import numpy as np

import moderngl
from ported._example import Example


class Fractal(Example):
    title = "Julia Set"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;
                out vec2 v_text;

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                    v_text = in_vert;
                }
            ''',
            fragment_shader='''
                #version 330

                in vec2 v_text;
                out vec4 f_color;

                uniform sampler2D Texture;
                uniform vec2 Seed;
                uniform int Iter;

                void main() {
                    vec2 c = Seed;
                    int i;

                    vec2 z = vec2(3.0 * v_text.x, 2.0 * v_text.y);

                    for (i = 0; i < Iter; i++) {
                        float x = (z.x * z.x - z.y * z.y) + c.x;
                        float y = (z.y * z.x + z.x * z.y) + c.y;

                        if ((x * x + y * y) > 4.0) {
                            break;
                        }

                        z.x = x;
                        z.y = y;
                    }

                    f_color = texture(Texture, vec2((i == Iter ? 0.0 : float(i)) / 100.0, 0.0));
                }
            ''',
        )

        self.seed = self.prog['Seed']
        self.iter = self.prog['Iter']

        self.texture = self.load_texture_2d('pal.png')

        vertices = np.array([-1.0, -1.0, -1.0, 1.0, 1.0, -1.0, 1.0, 1.0])

        self.vbo = self.ctx.buffer(vertices.astype('f4'))
        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, 'in_vert')

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)

        self.seed.value = (-0.8, 0.156)
        self.iter.value = 100

        self.texture.use()
        self.vao.render(moderngl.TRIANGLE_STRIP)


if __name__ == '__main__':
    Fractal.run()



"""
Compute shader renders a 32 x 32 grid to a 512, 512 texture
"""
import moderngl as mgl
from ported._example import Example
from moderngl_window import geometry


class RenderTextureCompute(Example):
    title = "Render Texture Using Compute Shader"
    gl_version = (4, 3)
    aspect_ratio = 1.0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # reference compute shader: http://wili.cc/blog/opengl-cs.html
        self.compute = self.ctx.compute_shader('''
            #version 430

            layout (local_size_x = 16, local_size_y = 16) in;
            // match the input texture format!
            layout(rgba8, location=0) writeonly uniform image2D destTex;

            uniform float time;

            void main() {
                // texel coordinate we are writing to
                ivec2 texelPos = ivec2(gl_GlobalInvocationID.xy);
                // Calculate 1.0 - distance from the center in each work group
                float local = 1.0 - length(vec2(ivec2(gl_LocalInvocationID.xy) - 8) / 8.0);
                // Wave covering the screen diagonally
                float global = sin(float(gl_WorkGroupID.x + gl_WorkGroupID.y) * 0.1 + time) / 2.0 + 0.5;
                imageStore(
                    destTex,
                    texelPos,
                    vec4(
                        local,
                        global,
                        0.0,
                        1.0
                    )
                );
            }
        ''')
        self.compute['destTex'] = 0

        # For rendering a simple textured quad
        self.quad_program = self.ctx.program(
            vertex_shader="""
            #version 330
            in vec3 in_position;
            in vec2 in_texcoord_0;
            out vec2 uv;

            void main() {
                gl_Position = vec4(in_position, 1.0);
                uv = in_texcoord_0;
            }
            """,
            fragment_shader="""
            #version 330
            uniform sampler2D texture0;
            out vec4 fragColor;
            in vec2 uv;

            void main() {
                fragColor = texture(texture0, uv);
            }
            """,
        )

        # RGB_8 texture
        self.texture = self.ctx.texture((256, 256), 4)
        self.texture.filter = mgl.NEAREST, mgl.NEAREST
        self.quad_fs = geometry.quad_fs()

    def render(self, time, frame_time):
        self.ctx.clear(0.3, 0.3, 0.3)

        w, h = self.texture.size
        gw, gh = 16, 16
        nx, ny, nz = int(w/gw), int(h/gh), 1

        try:
            self.compute['time'] = time
        except Exception:
            pass
        # Automatically binds as a GL_R32F / r32f (read from the texture)
        self.texture.bind_to_image(0, read=False, write=True)
        self.compute.run(nx, ny, nz)

        # Render texture
        self.texture.use(location=0)
        self.quad_fs.render(self.quad_program)


if __name__ == '__main__':
    RenderTextureCompute.run()



import wx
import skia
import moderngl
from wx import glcanvas

GL_RGBA8 = 0x8058


class DrawCanvas(glcanvas.GLCanvas):
    def __init__(self, parent, size):
        glcanvas.GLCanvas.__init__(self, parent, -1, size=size)

        self.size = None
        self.init = False
        self.ctx = None
        self.glcanvas = glcanvas.GLContext(self)
        self.size = (800, 800)

        self.x_pos = 0
        self.y_pos = 0

        self.Bind(wx.EVT_SIZE, self.OnSize)
        self.Bind(wx.EVT_PAINT, self.OnPaint)

    def OnEraseBackground(self, event):
        pass  # Do nothing, to avoid flashing on MSW.

    def OnSize(self, event):
        wx.CallAfter(self.DoSetViewport)
        event.Skip()

    def DoSetViewport(self):
        # Set the viewport size
        self.size = self.GetClientSize()
        self.SetCurrent(self.glcanvas)
        if not self.ctx is None:
            self.ctx.set_viewport(0, 0, self.Size.width, self.Size.height)

    def InitGL(self):
        # Initialize the skia context with the moderngl context
        self.ctx = moderngl.create_context()
        context = skia.GrDirectContext.MakeGL()
        backend_render_target = skia.GrBackendRenderTarget(
            self.size[0],
            self.size[1],
            0,  # sampleCnt
            0,  # stencilBits
            skia.GrGLFramebufferInfo(0, GL_RGBA8))
        self.surface = skia.Surface.MakeFromBackendRenderTarget(
            context, backend_render_target, skia.kBottomLeft_GrSurfaceOrigin,
            skia.kRGBA_8888_ColorType, skia.ColorSpace.MakeSRGB())
        self.canvas = self.surface.getCanvas()
        
    def OnPaint(self, event):
        dc = wx.PaintDC(self)
        self.SetCurrent(self.glcanvas)
        if not self.init:
            self.InitGL()
            self.init = True
        self.OnDraw()

    def OnDraw(self):
        self.SetContextViewport(0, 0, self.Size.width, self.Size.height)
        self.ctx.clear(255.0, 255.0, 255.0, 0.0)
        self.DrawContext()
        self.SwapBuffers()

    def DrawContext(self):
        # Do your drawing with skia here

        paint = skia.Paint(Color=skia.ColorGREEN) 
        paint.setAntiAlias(True)
        
        self.canvas.drawCircle(self.x_pos, self.y_pos, 40, paint)
        self.canvas.drawSimpleText("Hello ModernGL, Skia & wxPython!", 90, 200, 
                                skia.Font(skia.Typeface("Arial"), 40), 
                                skia.Paint(Color=skia.ColorBLUE))
        
        self.surface.flushAndSubmit()

    def SetContextViewport(self, x, y, width, height):
        self.ctx.viewport = (x, y, width, height)

    def SetXPos(self, x_pos):
        self.x_pos = x_pos
        self.Refresh(False)

    def SetYPos(self, y_pos):
        self.y_pos = y_pos
        self.Refresh(False)


class Frame(wx.Frame): 
    def __init__(self, parent, title): 
        super(Frame, self).__init__(parent, title=title, size=(900, 800))  

        self.SetBackgroundColour(wx.Colour("#eee"))

        main_sz = wx.BoxSizer(wx.HORIZONTAL)

        sz = wx.BoxSizer(wx.VERTICAL)
        panel = wx.Panel(self)
        panel.SetBackgroundColour(wx.Colour("#eee"))

        self.canvas = DrawCanvas(panel, size=(800, 800))

        self.slider_y = wx.Slider(self, 100, 25, 1, 800, size=(50, -1),
                            style=wx.SL_VERTICAL)
        self.slider_x = wx.Slider(panel, 100, 25, 1, 800, size=(50, -1), 
                                  style=wx.SL_HORIZONTAL)

        panel.SetSizer(sz)

        sz.Add(self.slider_x, 0, flag=wx.EXPAND|wx.ALL)
        sz.Add(self.canvas, 0, flag=wx.EXPAND|wx.ALL)

        main_sz.Add(self.slider_y, 0, flag=wx.EXPAND|wx.ALL)
        main_sz.Add(panel, 0, flag=wx.EXPAND|wx.ALL)

        self.SetSizer(main_sz)

        self.slider_x.Bind(wx.EVT_SLIDER, self.OnChangeX)
        self.slider_y.Bind(wx.EVT_SLIDER, self.OnChangeY)

        self.Center()

    def OnChangeX(self, event):
        self.canvas.SetXPos(self.slider_x.GetValue())
        event.Skip()

    def OnChangeY(self, event):
        self.canvas.SetYPos(self.slider_y.GetValue())
        event.Skip()
		
ex = wx.App() 
win = Frame(None, "ModernGL + Skia Python + wxPython") 
win.Show(True)
ex.MainLoop()



import unittest

import numpy as np
import numpy.testing as npt

import moderngl

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_1(self):
        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;
                out vec2 v_text;

                void main() {
                    v_text = in_vert / 2.0 + 0.5;
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                uniform sampler2D Texture;

                in vec2 v_text;
                out vec4 f_color;

                void main() {
                    f_color = texture(Texture, v_text);
                }
            ''',
        )

        pixels = np.tile([0.0, 0.5, 1.0], 4 * 4).astype('f2')
        self.tex = self.ctx.texture((4, 4), 3, pixels, dtype='f2')

        self.vbo = self.ctx.buffer(np.array([-1, -1, -1, 1, 1, -1, 1, 1], dtype='f4'))
        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, 'in_vert')
        self.fbo = self.ctx.simple_framebuffer((4, 4))

        self.fbo.use()
        self.fbo.clear()
        self.tex.use()
        self.vao.render(moderngl.TRIANGLE_STRIP)
        res = np.frombuffer(self.fbo.read(dtype='f2'), dtype='f2')
        npt.assert_almost_equal(res, pixels, decimal=2)


if __name__ == '__main__':
    unittest.main()



from context_manager import ContextManager


class RenderToTexture:
    def __init__(self, size=(256, 256), components=4, samples=4):
        self.ctx = ContextManager.get_default_context()
        self.texture = self.ctx.texture(size, components=4)
        self.fbo1 = self.ctx.simple_framebuffer(size, samples=samples)
        self.fbo2 = self.ctx.framebuffer(self.texture)
        self.scope = self.ctx.scope(self.fbo1)

    def __enter__(self):
        self.scope.__enter__()

    def __exit__(self, *args):
        self.scope.__exit__()
        self.ctx.copy_framebuffer(self.fbo2, self.fbo1)



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import GLWindow
# import ModernGL

# # Window & Context

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# vert = ctx.vertex_shader('''
#     #version 330

#     in vec2 vert;
#     out vec2 tex;

#     void main() {
#         gl_Position = vec4(vert, 0.0, 1.0);
#         tex = vert / 2.0 + vec2(0.5, 0.5);
#     }
# ''')

# frag = ctx.fragment_shader('''
#     #version 330

#     in vec2 tex;
#     out vec4 color;

#     uniform vec2 center;
#     uniform int iter;

#     void main() {
#         vec2 z = vec2(5.0 * (tex.x - 0.5), 3.0 * (tex.y - 0.5));
#         vec2 c = center;

#         int i;
#         for(i = 0; i < iter; i++) {
#             vec2 v = vec2(
#                 (z.x * z.x - z.y * z.y) + c.x,
#                 (z.y * z.x + z.x * z.y) + c.y
#             );
#             if (dot(v, v) > 4.0) break;
#             z = v;
#         }

#         float cm = fract((i == iter ? 0.0 : float(i)) * 10 / iter);
#         color = vec4(
#             fract(cm + 0.0 / 3.0),
#             fract(cm + 1.0 / 3.0),
#             fract(cm + 2.0 / 3.0),
#             1.0
#         );
#     }
# ''')

# prog = ctx.program(vert, frag])

# vbo = ctx.buffer(struct.pack('8f', -1.0, -1.0, -1.0, 1.0, 1.0, -1.0, 1.0, 1.0))
# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])

# prog.uniforms['iter'].value = 100

# x, y = (0.49, 0.32)

# wnd.grab_mouse(True)

# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)
#     mx, my = wnd.mouse_delta
#     x -= mx / 100
#     y -= my / 100

#     prog.uniforms['center'].value = (y, x)
#     vao.render(ModernGL.TRIANGLE_STRIP)



import random
import math
import struct
import unittest

import moderngl
import numpy as np

from common import get_context


class TestScissor(unittest.TestCase):
    """
    Test blending and blend equations with float32 framebuffer.
    Emulate blend and equation values in python and compare with moderngl.
    """
    @classmethod
    def setUpClass(cls):
        """Set up program, vbo and fbo"""
        cls.ctx = get_context()
        cls.prog = cls.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 fragColor;
                uniform vec4 color;

                void main() {
                    fragColor = color;
                }
            ''',
        )
        quad = [
            -1.0,  1.0,
            -1.0, -1.0,
            1.0, 1.0,
            1.0, -1.0,
        ]
        vbo = cls.ctx.buffer(np.array(quad, dtype='f4').tobytes())
        cls.vao = cls.ctx.simple_vertex_array(cls.prog, vbo, 'in_vert')

    def create_fbo(self, size, components=3):
        return self.ctx.framebuffer(
            color_attachments=[self.ctx.texture(size, components)],
        )

    def test_default_value(self):
        """Ensure default value of the scissor matches the framebuffer size"""
        size = (16, 32)
        fbo = self.create_fbo(size)
        self.assertEqual(fbo.scissor, (0, 0, *size))

    def test_reset_scissor(self):
        """We should be able to reset scissor with `None`"""
        size = (16, 32)
        fbo = self.create_fbo(size)
        new_val = 2, 2, 14, 30

        fbo.scissor = new_val
        self.assertEqual(fbo.scissor, new_val)

        fbo.scissor = None
        self.assertEqual(fbo.scissor, (0, 0, *size))

    def test_render(self):
        """Render different color to the four corners of the fbo using scissor"""
        size = (4, 4)
        fbo = self.create_fbo(size)
        fbo.use()

        # upper left (red)
        fbo.scissor = 0, 2, 2, 2
        self.prog['color'].value = 1.0, 0.0, 0.0, 0.0
        self.vao.render(mode=moderngl.TRIANGLE_STRIP)

        # upper right (green)
        fbo.scissor = 2, 2, 2, 2
        self.prog['color'].value = 0.0, 1.0, 0.0, 0.0
        self.vao.render(mode=moderngl.TRIANGLE_STRIP)

        # lower left (green)
        fbo.scissor = 0, 0, 2, 2
        self.prog['color'].value = 0.0, 0.0, 1.0, 0.0
        self.vao.render(mode=moderngl.TRIANGLE_STRIP)

        # lower right (green)
        fbo.scissor = 2, 0, 2, 2
        self.prog['color'].value = 1.0, 1.0, 1.0, 1.0
        self.vao.render(mode=moderngl.TRIANGLE_STRIP)

        data = fbo.read()
        # Note that the data is flipped horizontally
        self.assertEqual(
            data,
            (b'\x00\x00\xff'b'\x00\x00\xff'b'\xff\xff\xff'b'\xff\xff\xff'
             b'\x00\x00\xff'b'\x00\x00\xff'b'\xff\xff\xff'b'\xff\xff\xff'
             b'\xff\x00\x00'b'\xff\x00\x00'b'\x00\xff\x00'b'\x00\xff\x00'
             b'\xff\x00\x00'b'\xff\x00\x00'b'\x00\xff\x00'b'\x00\xff\x00')
        )

    def test_clear(self):
        """Ensure fbo clearing works with scissoring"""
        size = (4, 4)
        fbo = self.create_fbo(size)
        fbo.use()

        # upper left (red)
        fbo.scissor = 0, 2, 2, 2
        fbo.clear(1.0, 0.0, 0.0, 0.0)

        # upper right (green)
        fbo.scissor = 2, 2, 2, 2
        fbo.clear(0.0, 1.0, 0.0, 0.0)

        # lower left (green)
        fbo.scissor = 0, 0, 2, 2
        fbo.clear(0.0, 0.0, 1.0, 0.0)

        # lower right (green)
        fbo.scissor = 2, 0, 2, 2
        fbo.clear(1.0, 1.0, 1.0, 1.0)

        data = fbo.read()
        # Note that the data is flipped horizontally
        self.assertEqual(
            data,
            (b'\x00\x00\xff'b'\x00\x00\xff'b'\xff\xff\xff'b'\xff\xff\xff'
             b'\x00\x00\xff'b'\x00\x00\xff'b'\xff\xff\xff'b'\xff\xff\xff'
             b'\xff\x00\x00'b'\xff\x00\x00'b'\x00\xff\x00'b'\x00\xff\x00'
             b'\xff\x00\x00'b'\xff\x00\x00'b'\x00\xff\x00'b'\x00\xff\x00')
        )

    def test_scissor_leak(self):
        """Make sure we don't leak scissor values to other fbos"""
        size = 2, 2
        fbo1 = self.create_fbo(size, components=4)
        fbo2 = self.create_fbo(size, components=4)

        fbo1.scissor = 0, 0, 1, 1
        fbo2.scissor = 1, 1, 1, 1

        fbo1.clear(color=(1.0, 0.0, 0.0, 0.0))
        fbo2.clear(color=(0.0, 1.0, 0.0, 0.0))

        self.assertEqual(fbo1.read(), b'\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
        self.assertEqual(fbo2.read(), b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x00')

    def test_clear_with_viewport(self):
        """Clearing with viewport take precense over scissor"""
        size = 2, 1
        fbo = self.create_fbo(size)
        fbo.scissor = 1, 0, 1, 1
        fbo.clear(color=(1.0, 1.0, 1.0, 1.0), viewport=(0, 0, 1, 1))
        self.assertEqual(fbo.read(), b'\xff\xff\xff\x00\x00\x00')



'''
    Renders 2 triangles that share the same head using index buffers (ibo)
'''

import numpy as np

from ported._example import Example


class IndexBuffer(Example):
    gl_version = (3, 3)
    title = "Index Buffer"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 f_color;

                void main() {
                    f_color = vec4(0.3, 0.5, 1.0, 1.0);
                }
            ''',
        )

        # 2 triangles sharing the head vertex (0,0)
        vertices = np.array([
            0.0, 0.0,

            -0.6, -0.8,
            0.6, -0.8,

            0.6, 0.8,
            -0.6, 0.8,
        ], dtype='f4')

        # Indices are given to specify the order of drawing
        indices = np.array([0, 1, 2, 0, 3, 4], dtype='i4')

        self.vbo = self.ctx.buffer(vertices)
        self.ibo = self.ctx.buffer(indices)

        vao_content = [
            # 2 floats are assigned to the 'in' variable named 'in_vert' in the shader code
            (self.vbo, '2f', 'in_vert')
        ]

        self.vao = self.ctx.vertex_array(self.prog, vao_content, self.ibo)

    def render(self, time: float, frame_time: float):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.vao.render()


if __name__ == '__main__':
    IndexBuffer.run()



"""
Render using empty Vertexbuffer.
Renders 100 triangels emitted by a geometry shaders.
In addition we test if instancing is working passing gl_InstanceID from the vertex shader.
"""
import moderngl

from ported._example import Example


class HelloWorld(Example):
    gl_version = (3, 3)
    title = "Empty Vertexbuffer"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330
                out int inst;
                void main() {
                    inst = gl_InstanceID;
                }
            ''',
            geometry_shader='''
                #version 330
                layout (points) in;
                layout (triangle_strip, max_vertices = 3) out;
                in int inst[1];
                void main() {
                    float x = float(gl_PrimitiveIDIn / 10) / 9 - 0.5 + inst[0] / 20.0;
                    float y = float(gl_PrimitiveIDIn % 10) / 9 - 0.5 + inst[0] / 20.0;
                    gl_Position = vec4(x - 0.03, y - 0.03, 0.0, 1.0);
                    EmitVertex();
                    gl_Position = vec4(x + 0.03, y - 0.03, 0.0, 1.0);
                    EmitVertex();
                    gl_Position = vec4(x, y + 0.03, 0.0, 1.0);
                    EmitVertex();
                    EndPrimitive();
                }
            ''',
            fragment_shader='''
                #version 330
                out vec4 f_color;
                void main() {
                    f_color = vec4(0.3, 0.5, 1.0, 1.0);
                }
            ''',
        )
        self.vao = self.ctx.vertex_array(self.prog, [])

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.vao.render(mode=moderngl.POINTS, vertices=100, instances=2)


if __name__ == '__main__':
    HelloWorld.run()



import struct

import ModernGL
from PyQt5 import QtOpenGL, QtWidgets


class QGLControllerWidget(QtOpenGL.QGLWidget):
    def __init__(self):
        fmt = QtOpenGL.QGLFormat()
        fmt.setVersion(3, 3)
        fmt.setProfile(QtOpenGL.QGLFormat.CoreProfile)
        fmt.setSampleBuffers(True)
        super(QGLControllerWidget, self).__init__(fmt, None)

    def initializeGL(self):
        self.ctx = ModernGL.create_context()

        prog = self.ctx.program(
            vertex_shader='''
    #version 330

    in vec3 vert;

    uniform float znear;
    uniform float zfar;
    uniform float fovy;
    uniform float ratio;

    uniform vec3 center;
    uniform vec3 eye;
    uniform vec3 up;

    mat4 perspective() {
        float zmul = (-2.0 * znear * zfar) / (zfar - znear);
        float ymul = 1.0 / tan(fovy * 3.14159265 / 360);
        float xmul = ymul / ratio;

        return mat4(
            xmul, 0.0, 0.0, 0.0,
            0.0, ymul, 0.0, 0.0,
            0.0, 0.0, -1.0, -1.0,
            0.0, 0.0, zmul, 0.0
        );
    }

    mat4 lookat() {
        vec3 forward = normalize(center - eye);
        vec3 side = normalize(cross(forward, up));
        vec3 upward = cross(side, forward);
        return mat4(
            side.x, upward.x, -forward.x, 0,
            side.y, upward.y, -forward.y, 0,
            side.z, upward.z, -forward.z, 0,
            -dot(eye, side), -dot(eye, upward), dot(eye, forward), 1
        );
    }

    void main() {
        gl_Position = perspective() * lookat() * vec4(vert, 1.0);
    }
'''),
            fragment_shader='''
    #version 330

    out vec4 color;

    void main() {
        color = vec4(0.04, 0.04, 0.04, 1.0);
    }
'''),
        ])

        self.znear = prog.uniforms['znear']
        self.znear.value = 0.1
        self.zfar = prog.uniforms['zfar']
        self.zfar.value = 1000.0
        self.ratio = prog.uniforms['ratio']
        self.ratio.value = self.width() / self.height()
        self.fovy = prog.uniforms['fovy']
        self.fovy.value = 60

        self.eye = prog.uniforms['eye']
        self.eye.value = (3, 3, 3)
        self.center = prog.uniforms['center']
        self.center.value = (0, 0, 0)
        self.up = prog.uniforms['up']
        self.up.value = (0, 0, 1)

        grid = bytes()

        for i in range(0, 65):
            grid += struct.pack('6f', i - 32, -32.0, 0.0, i - 32, 32.0, 0.0)
            grid += struct.pack('6f', -32.0, i - 32, 0.0, 32.0, i - 32, 0.0)

        self.vbo = self.ctx.buffer(grid)
        self.vao = self.ctx.simple_vertex_array(prog, self.vbo, ['vert'])

    def paintGL(self):

        self.ctx.viewport = (0, 0, self.width(), self.height())
        self.ratio.value = self.width() / self.height()
        self.ctx.clear(0.9, 0.9, 0.9)
        self.vao.render(ModernGL.LINES, 65 * 4)
        self.ctx.finish()


app = QtWidgets.QApplication([])
window = QGLControllerWidget()
window.move(QtWidgets.QDesktopWidget().rect().center() - window.rect().center())
window.show()
app.exec_()



import moderngl

# NOTE: According to the spec we need a framebuffer bound at all times
#       to not break tests. Making sure a context always has a framebuffer
#       bound ensures tests works across platforms.
# OpenGL 4.4 Core Specification - 9.4.4 Effects of Framebuffer Completeness on Framebuffer Operations
# 
# A GL_INVALID_FRAMEBUFFER_OPERATION error is generated by attempts to
# render to or read from a framebuffer which is not framebuffer complete.
# This error is generated regardless of whether fragments are actually
# read from or written to the framebuffer. For example, it is generated
# when a rendering command is called and the framebuffer is incomplete,
# even if GL_RASTERIZER_DISCARD is enabled.

# structure {
#     'context_330': {
#         'ctx': Context
#         'fbo': Framebuffer
#     },
# }
_static = {}


def get_context(require=330) -> moderngl.Context:
    """Gets a context, cache it and activate it"""
    key = 'context_{}'.format(require)

    entry = _static.get(key)

    if entry is None:
        try:
            ctx = moderngl.create_context(require=require, standalone=True)
            entry = {'ctx': ctx, 'fbo': ctx.simple_framebuffer((100, 100), 4)}
            _static[key] = entry
        except Exception as ex:
            print(ex)
            return None

    entry['fbo'].use()
    entry['fbo'].clear()
    entry['ctx'].__enter__()
    return entry['ctx']
 


import unittest

from common import get_context
import moderngl


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

    def test_program(self):
        program = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform vec2 pos;
                uniform float scale;

                in vec2 vert;
                out vec2 v_vert;

                void main() {
                    gl_Position = vec4(pos + vert * scale, 0.0, 1.0);
                    v_vert = vert;
                }
            ''',
            fragment_shader='''
                #version 330

                in vec2 v_vert;
                out vec4 color;

                void main() {
                    color = vec4(v_vert, 0.0, 1.0);
                }
            ''',
        )

        self.assertIn('vert', program)
        self.assertIn('pos', program)
        self.assertIn('scale', program)
        self.assertIs(program.is_transform, False)

        self.assertIsInstance(program['vert'], moderngl.Attribute)
        self.assertIsInstance(program['pos'], moderngl.Uniform)
        self.assertIsInstance(program['scale'], moderngl.Uniform)

    def test_geo_input_output_primitive(self):
        """Test detection of all geo shader input/output primitives types"""
        input_types = [
            ("points", moderngl.POINTS),
            ("lines", moderngl.LINES),
            ("lines_adjacency", moderngl.LINES_ADJACENCY),
            ("triangles", moderngl.TRIANGLES),
            ("triangles_adjacency", moderngl.TRIANGLES_ADJACENCY),
        ]
        output_types = [
            ("points", moderngl.POINTS),
            ("line_strip", moderngl.LINES),
            ("triangle_strip", moderngl.TRIANGLES),
        ]

        for out_name, out_type in output_types:
            for in_name, in_type in input_types:
                p = self.ctx.program(
                    vertex_shader="""
                    #version 330
                    void main() {
                        gl_Position = vec4(1.0);
                    }
                    """,
                    geometry_shader=(
                        "#version 330\n"
                        f"layout({in_name}) in;\n"
                        f"layout({out_name}, max_vertices=6) out;\n"
                        "void main() {\n"
                        "    gl_Position = vec4(1.0);\n"
                        "    EmitVertex();\n"
                        "    EndPrimitive();\n"
                        "}\n"
                    ),
                )
                self.assertEqual(p.geometry_input, in_type)
                self.assertEqual(p.geometry_output, out_type, msg=f"input: {in_name}, output: {out_name}")


if __name__ == '__main__':
    unittest.main()



import struct
from typing import Any


class Attribute:
    """This class represents a program attribute."""

    def __init__(self):
        self._location: int = None
        self._array_length: int = None
        self._dimension: int = None
        self._shape: int = None
        self._name: str = None
        self.extra: Any = None  #: Attribute for storing user defined objects

    def __repr__(self) -> str:
        return '<Attribute: %d>' % self._location

    @property
    def mglo(self):
        return self

    @property
    def location(self) -> int:
        """
        int: The location of the attribute.

        The result of the glGetAttribLocation.
        """
        return self._location

    @property
    def array_length(self) -> int:
        """
        int: If the attribute is an array the array_length.

        is the length of the array otherwise `1`.
        """
        return self._array_length

    @property
    def dimension(self) -> int:
        """
        int: The attribute dimension.

        +-----------+-----------+
        | GLSL type | dimension |
        +===========+===========+
        | int       | 1         |
        +-----------+-----------+
        | ivec2     | 2         |
        +-----------+-----------+
        | ivec3     | 3         |
        +-----------+-----------+
        | ivec4     | 4         |
        +-----------+-----------+
        | uint      | 1         |
        +-----------+-----------+
        | uvec2     | 2         |
        +-----------+-----------+
        | uvec3     | 3         |
        +-----------+-----------+
        | uvec4     | 4         |
        +-----------+-----------+
        | float     | 1         |
        +-----------+-----------+
        | vec2      | 2         |
        +-----------+-----------+
        | vec3      | 3         |
        +-----------+-----------+
        | vec4      | 4         |
        +-----------+-----------+
        | double    | 1         |
        +-----------+-----------+
        | dvec2     | 2         |
        +-----------+-----------+
        | dvec3     | 3         |
        +-----------+-----------+
        | dvec4     | 4         |
        +-----------+-----------+
        | mat2      | 4         |
        +-----------+-----------+
        | mat2x3    | 6         |
        +-----------+-----------+
        | mat2x4    | 8         |
        +-----------+-----------+
        | mat3x2    | 6         |
        +-----------+-----------+
        | mat3      | 9         |
        +-----------+-----------+
        | mat3x4    | 12        |
        +-----------+-----------+
        | mat4x2    | 8         |
        +-----------+-----------+
        | mat4x3    | 12        |
        +-----------+-----------+
        | mat4      | 16        |
        +-----------+-----------+
        | dmat2     | 4         |
        +-----------+-----------+
        | dmat2x3   | 6         |
        +-----------+-----------+
        | dmat2x4   | 8         |
        +-----------+-----------+
        | dmat3x2   | 6         |
        +-----------+-----------+
        | dmat3     | 9         |
        +-----------+-----------+
        | dmat3x4   | 12        |
        +-----------+-----------+
        | dmat4x2   | 8         |
        +-----------+-----------+
        | dmat4x3   | 12        |
        +-----------+-----------+
        | dmat4     | 16        |
        +-----------+-----------+
        """
        return self._dimension

    @property
    def shape(self) -> int:
        """
        str: The shape is a single character, representing the scalar type of the attribute.

        +---------+--------------------------+
        | shape   | GLSL types               |
        +=========+==========================+
        | ``'i'`` | int                      |
        |         +--------------------------+
        |         | ivec2 ivec3 ivec4        |
        +---------+--------------------------+
        | ``'I'`` | uint                     |
        |         +--------------------------+
        |         | uvec2 uvec3 uvec4        |
        +---------+--------------------------+
        | ``'f'`` | float                    |
        |         +--------------------------+
        |         | vec2 vec3 vec4           |
        |         +--------------------------+
        |         | mat2 mat3 mat4           |
        |         +--------------------------+
        |         | mat2x3 mat2x4 mat3x4     |
        |         | mat4x2 mat4x2 mat4x3     |
        +---------+--------------------------+
        | ``'d'`` | double                   |
        |         +--------------------------+
        |         | dvec2 dvec3 dvec4        |
        |         +--------------------------+
        |         | dmat2 dmat3 dmat4        |
        |         +--------------------------+
        |         | dmat2x3 dmat2x4 dmat3x4  |
        |         | dmat4x2 dmat4x2 dmat4x3  |
        +---------+--------------------------+
        """
        return self._shape

    @property
    def name(self) -> str:
        """
        str: The attribute name.

        The name will be filtered to have no array syntax on it's end.
        Attribute name without ``'[0]'`` ending if any.
        """
        return self._name


class Uniform:
    """
    A uniform is a global GLSL variable declared with the "uniform" storage qualifier.

    These act as parameters that the user of a shader program can pass to that program.

    In ModernGL, Uniforms can be accessed using :py:meth:`Program.__getitem__`
    or :py:meth:`Program.__iter__`
    """

    def __init__(self):
        self._program_obj = None
        self._type = None
        self._fmt = None
        self._location = None
        self._array_length = None
        self._element_size = None
        self._dimension = None
        self._name = None
        self._matrix = None
        self._ctx = None
        self.extra = None  #: Any - Attribute for storing user defined objects

    def __repr__(self) -> str:
        return '<Uniform: %d>' % self._location

    @property
    def mglo(self):
        return self

    @property
    def location(self) -> int:
        """
        int: The location of the uniform.

        The location holds the value returned by the glGetUniformLocation.
        To set the value of the uniform use the :py:attr:`value` instead.
        """
        return self._location

    @property
    def dimension(self) -> int:
        """
        int: The dimension of the uniform.

        +-----------------+-----------+
        | GLSL type       | dimension |
        +=================+===========+
        | sampler2D       | 1         |
        +-----------------+-----------+
        | sampler2DCube   | 1         |
        +-----------------+-----------+
        | sampler2DShadow | 1         |
        +-----------------+-----------+
        | bool            | 1         |
        +-----------------+-----------+
        | bvec2           | 2         |
        +-----------------+-----------+
        | bvec3           | 3         |
        +-----------------+-----------+
        | bvec4           | 4         |
        +-----------------+-----------+
        | int             | 1         |
        +-----------------+-----------+
        | ivec2           | 2         |
        +-----------------+-----------+
        | ivec3           | 3         |
        +-----------------+-----------+
        | ivec4           | 4         |
        +-----------------+-----------+
        | uint            | 1         |
        +-----------------+-----------+
        | uvec2           | 2         |
        +-----------------+-----------+
        | uvec3           | 3         |
        +-----------------+-----------+
        | uvec4           | 4         |
        +-----------------+-----------+
        | float           | 1         |
        +-----------------+-----------+
        | vec2            | 2         |
        +-----------------+-----------+
        | vec3            | 3         |
        +-----------------+-----------+
        | vec4            | 4         |
        +-----------------+-----------+
        | double          | 1         |
        +-----------------+-----------+
        | dvec2           | 2         |
        +-----------------+-----------+
        | dvec3           | 3         |
        +-----------------+-----------+
        | dvec4           | 4         |
        +-----------------+-----------+
        | mat2            | 4         |
        +-----------------+-----------+
        | mat2x3          | 6         |
        +-----------------+-----------+
        | mat2x4          | 8         |
        +-----------------+-----------+
        | mat3x2          | 6         |
        +-----------------+-----------+
        | mat3            | 9         |
        +-----------------+-----------+
        | mat3x4          | 12        |
        +-----------------+-----------+
        | mat4x2          | 8         |
        +-----------------+-----------+
        | mat4x3          | 12        |
        +-----------------+-----------+
        | mat4            | 16        |
        +-----------------+-----------+
        | dmat2           | 4         |
        +-----------------+-----------+
        | dmat2x3         | 6         |
        +-----------------+-----------+
        | dmat2x4         | 8         |
        +-----------------+-----------+
        | dmat3x2         | 6         |
        +-----------------+-----------+
        | dmat3           | 9         |
        +-----------------+-----------+
        | dmat3x4         | 12        |
        +-----------------+-----------+
        | dmat4x2         | 8         |
        +-----------------+-----------+
        | dmat4x3         | 12        |
        +-----------------+-----------+
        | dmat4           | 16        |
        +-----------------+-----------+
        """
        return self._dimension

    @property
    def array_length(self) -> int:
        """
        int: The length of the array of the uniform.

        The array_length is `1` for non array uniforms.
        """
        return self._array_length

    @property
    def name(self) -> str:
        """
        str: The name of the uniform.

        The name does not contain leading `[0]`.
        The name may contain `[ ]` when the uniform is part of a struct.
        """
        return self._name

    @property
    def value(self) -> Any:
        """
        The value of the uniform.

        Reading the value of the uniform may force the GPU to sync.

        The value must be a tuple for non array uniforms.
        The value must be a list of tuples for array uniforms.
        """
        data = self.read()
        if self._array_length > 1:
            if self._dimension > 1:
                return [
                    struct.unpack(self._fmt, data[i * self._element_size : i * self._element_size + self._element_size])
                    for i in range(self._array_length)
                ]
            else:
                return [
                    struct.unpack(self._fmt, data[i * self._element_size : i * self._element_size + self._element_size])[0]
                    for i in range(self._array_length)
                ]
        elif self._dimension > 1:
            return struct.unpack(self._fmt, data)
        else:
            return struct.unpack(self._fmt, data)[0]

    @value.setter
    def value(self, value: Any) -> None:
        if self._array_length > 1:
            if self._dimension > 1:
                data = b''.join(struct.pack(self._fmt, *row) for row in value)
            else:
                data = b''.join(struct.pack(self._fmt, item) for item in value)
        elif self._dimension > 1:
            data = struct.pack(self._fmt, *value)
        else:
            data = struct.pack(self._fmt, value)
        self.write(data)

    def read(self) -> bytes:
        """Read the value of the uniform."""
        return self._ctx._read_uniform(self._program_obj, self._location, self._type, self._array_length, self._element_size)

    def write(self, data: Any) -> None:
        """Write the value of the uniform."""
        self._ctx._write_uniform(self._program_obj, self._location, self._type, self._array_length, data)


class UniformBlock:
    """Uniform Block metadata."""

    def __init__(self):
        self._program_obj = None
        self._index = None
        self._size = None
        self._name = None
        self._ctx = None
        self.extra = None  #: Any - Attribute for storing user defined objects

    def __repr__(self):
        return '<UniformBlock: %d>' % self._index

    @property
    def mglo(self):
        return self

    @property
    def binding(self) -> int:
        """int: The binding of the uniform block."""
        return self._ctx._get_ubo_binding(self._program_obj, self._index)

    @binding.setter
    def binding(self, binding: int) -> None:
        self._ctx._set_ubo_binding(self._program_obj, self._index, binding)

    @property
    def value(self) -> int:
        """int: The value of the uniform block."""
        return self._ctx._get_ubo_binding(self._program_obj, self._index)

    @value.setter
    def value(self, value: int) -> None:
        self._ctx._set_ubo_binding(self._program_obj, self._index, value)

    @property
    def name(self) -> str:
        """str: The name of the uniform block."""
        return self._name

    @property
    def index(self) -> int:
        """int: The index of the uniform block."""
        return self._index

    @property
    def size(self) -> int:
        """int: The size of the uniform block."""
        return self._size



class Subroutine:
    """This class represents a program subroutine."""

    def __init__(self):
        self._index = None
        self._name = None
        self.extra = None  #: Any - Attribute for storing user defined objects

    def __repr__(self) -> str:
        return '<Subroutine: %d>' % self._index

    @property
    def mglo(self):
        return self

    @property
    def index(self) -> int:
        """int: The index of the subroutine."""
        return self._index

    @property
    def name(self) -> str:
        """str: The name of the subroutine."""
        return self._name


class Varying:
    """This class represents a program varying."""

    def __init__(self):
        self._number = None
        self._array_length = None
        self._dimension = None
        self._name = None
        self.extra: Any = None  #: Attribute for storing user defined objects

    def __repr__(self) -> str:
        return '<Varying: %d>' % self.number

    @property
    def mglo(self):
        return self

    @property
    def number(self) -> int:
        """int: The number of the varying."""
        return self._number

    @property
    def name(self) -> str:
        """str: The name of the varying."""
        return self._name


class Error(Exception):
    """Generic moderngl error."""
    pass


ATTRIBUTE_LOOKUP_TABLE = {
    0x1404: (1, 0x1404, 1, 1, False, 'i'),
    0x8b53: (2, 0x1404, 1, 2, False, 'i'),
    0x8b54: (3, 0x1404, 1, 3, False, 'i'),
    0x8b55: (4, 0x1404, 1, 4, False, 'i'),
    0x1405: (1, 0x1405, 1, 1, False, 'i'),
    0x8dc6: (2, 0x1405, 1, 2, False, 'i'),
    0x8dc7: (3, 0x1405, 1, 3, False, 'i'),
    0x8dc8: (4, 0x1405, 1, 4, False, 'i'),
    0x1406: (1, 0x1406, 1, 1, True, 'f'),
    0x8b50: (2, 0x1406, 1, 2, True, 'f'),
    0x8b51: (3, 0x1406, 1, 3, True, 'f'),
    0x8b52: (4, 0x1406, 1, 4, True, 'f'),
    0x140a: (1, 0x140a, 1, 1, False, 'd'),
    0x8ffc: (2, 0x140a, 1, 2, False, 'd'),
    0x8ffd: (3, 0x140a, 1, 3, False, 'd'),
    0x8ffe: (4, 0x140a, 1, 4, False, 'd'),
    0x8b5a: (4, 0x1406, 2, 2, True, 'f'),
    0x8b65: (6, 0x1406, 2, 3, True, 'f'),
    0x8b66: (8, 0x1406, 2, 4, True, 'f'),
    0x8b67: (6, 0x1406, 3, 2, True, 'f'),
    0x8b5b: (9, 0x1406, 3, 3, True, 'f'),
    0x8b68: (12, 0x1406, 3, 4, True, 'f'),
    0x8b69: (8, 0x1406, 4, 2, True, 'f'),
    0x8b6a: (12, 0x1406, 4, 3, True, 'f'),
    0x8b5c: (16, 0x1406, 4, 4, True, 'f'),
    0x8f46: (4, 0x140a, 2, 2, False, 'd'),
    0x8f49: (6, 0x140a, 2, 3, False, 'd'),
    0x8f4a: (8, 0x140a, 2, 4, False, 'd'),
    0x8f4b: (6, 0x140a, 3, 2, False, 'd'),
    0x8f47: (9, 0x140a, 3, 3, False, 'd'),
    0x8f4c: (12, 0x140a, 3, 4, False, 'd'),
    0x8f4d: (8, 0x140a, 4, 2, False, 'd'),
    0x8f4e: (12, 0x140a, 4, 3, False, 'd'),
    0x8f48: (16, 0x140a, 4, 4, False, 'd'),
}

UNIFORM_LOOKUP_TABLE = {
    0x8B56: (False, 1, 4, '1i'),  # GL_BOOL
    0x8B57: (False, 2, 8, '2i'),  # GL_BOOL_VEC2
    0x8B58: (False, 3, 12, '3i'),  # GL_BOOL_VEC3
    0x8B59: (False, 4, 16, '4i'),  # GL_BOOL_VEC4
    0x1404: (False, 1, 4, '1i'),  # GL_INT
    0x8B53: (False, 2, 8, '2i'),  # GL_INT_VEC2
    0x8B54: (False, 3, 12, '3i'),  # GL_INT_VEC3
    0x8B55: (False, 4, 16, '4i'),  # GL_INT_VEC4
    0x1405: (False, 1, 4, '1I'),  # GL_UNSIGNED_INT
    0x8DC6: (False, 2, 8, '2I'),  # GL_UNSIGNED_INT_VEC2
    0x8DC7: (False, 3, 12, '3I'),  # GL_UNSIGNED_INT_VEC3
    0x8DC8: (False, 4, 16, '4I'),  # GL_UNSIGNED_INT_VEC4
    0x1406: (False, 1, 4, '1f'),  # GL_FLOAT
    0x8B50: (False, 2, 8, '2f'),  # GL_FLOAT_VEC2
    0x8B51: (False, 3, 12, '3f'),  # GL_FLOAT_VEC3
    0x8B52: (False, 4, 16, '4f'),  # GL_FLOAT_VEC4
    0x140A: (False, 1, 8, '1d'),  # GL_DOUBLE
    0x8FFC: (False, 2, 16, '2d'),  # GL_DOUBLE_VEC2
    0x8FFD: (False, 3, 24, '3d'),  # GL_DOUBLE_VEC3
    0x8FFE: (False, 4, 32, '4d'),  # GL_DOUBLE_VEC4
    0x8B5D: (False, 1, 4, '1i'),  # GL_SAMPLER_1D
    0x8DC0: (False, 1, 4, '1i'),  # GL_SAMPLER_1D_ARRAY
    0x8DC9: (False, 1, 4, '1i'),  # GL_INT_SAMPLER_1D
    0x8DCE: (False, 1, 4, '1i'),  # GL_INT_SAMPLER_1D_ARRAY
    0x8B5E: (False, 1, 4, '1i'),  # GL_SAMPLER_2D
    0x8DCA: (False, 1, 4, '1i'),  # GL_INT_SAMPLER_2D
    0x8DD2: (False, 1, 4, '1i'),  # GL_UNSIGNED_INT_SAMPLER_2D
    0x8DC1: (False, 1, 4, '1i'),  # GL_SAMPLER_2D_ARRAY
    0x8DCF: (False, 1, 4, '1i'),  # GL_INT_SAMPLER_2D_ARRAY
    0x8DD7: (False, 1, 4, '1i'),  # GL_UNSIGNED_INT_SAMPLER_2D_ARRAY
    0x8B5F: (False, 1, 4, '1i'),  # GL_SAMPLER_3D
    0x8DCB: (False, 1, 4, '1i'),  # GL_INT_SAMPLER_3D
    0x8DD3: (False, 1, 4, '1i'),  # GL_UNSIGNED_INT_SAMPLER_3D
    0x8B62: (False, 1, 4, '1i'),  # GL_SAMPLER_2D_SHADOW
    0x9108: (False, 1, 4, '1i'),  # GL_SAMPLER_2D_MULTISAMPLE
    0x9109: (False, 1, 4, '1i'),  # GL_INT_SAMPLER_2D_MULTISAMPLE
    0x910A: (False, 1, 4, '1i'),  # GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE
    0x910B: (False, 1, 4, '1i'),  # GL_SAMPLER_2D_MULTISAMPLE_ARRAY
    0x910C: (False, 1, 4, '1i'),  # GL_INT_SAMPLER_2D_MULTISAMPLE_ARRAY
    0x910D: (False, 1, 4, '1i'),  # GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE_ARRAY
    0x8B60: (False, 1, 4, '1i'),  # GL_SAMPLER_CUBE
    0x8DCC: (False, 1, 4, '1i'),  # GL_INT_SAMPLER_CUBE
    0x8DD4: (False, 1, 4, '1i'),  # GL_UNSIGNED_INT_SAMPLER_CUBE
    0x904D: (False, 1, 4, '1i'),  # GL_IMAGE_2D
    0x8B5A: (True, 4, 16, '4f'),  # GL_FLOAT_MAT2
    0x8B65: (True, 6, 24, '6f'),  # GL_FLOAT_MAT2x3
    0x8B66: (True, 8, 32, '8f'),  # GL_FLOAT_MAT2x4
    0x8B67: (True, 6, 24, '6f'),  # GL_FLOAT_MAT3x2
    0x8B5B: (True, 9, 36, '9f'),  # GL_FLOAT_MAT3
    0x8B68: (True, 12, 48, '12f'),  # GL_FLOAT_MAT3x4
    0x8B69: (True, 8, 32, '8f'),  # GL_FLOAT_MAT4x2
    0x8B6A: (True, 12, 48, '12f'),  # GL_FLOAT_MAT4x3
    0x8B5C: (True, 16, 64, '16f'),  # GL_FLOAT_MAT4
    0x8F46: (True, 4, 32, '4d'),  # GL_DOUBLE_MAT2
    0x8F49: (True, 6, 48, '6d'),  # GL_DOUBLE_MAT2x3
    0x8F4A: (True, 8, 64, '8d'),  # GL_DOUBLE_MAT2x4
    0x8F4B: (True, 6, 48, '6d'),  # GL_DOUBLE_MAT3x2
    0x8F47: (True, 9, 72, '9d'),  # GL_DOUBLE_MAT3
    0x8F4C: (True, 12, 96, '12d'),  # GL_DOUBLE_MAT3x4
    0x8F4D: (True, 8, 64, '8d'),  # GL_DOUBLE_MAT4x2
    0x8F4E: (True, 12, 96, '12d'),  # GL_DOUBLE_MAT4x3
    0x8F48: (True, 16, 128, '16d'),  # GL_DOUBLE_MAT4
}


def make_attribute(name, gl_type, program_obj, location, array_length):
    tmp = ATTRIBUTE_LOOKUP_TABLE.get(gl_type, (1, 0, 1, 1, False, '?'))
    dimension, scalar_type, rows_length, row_length, normalizable, shape = tmp
    rows_length *= array_length
    res = Attribute()
    res._type = gl_type
    res._program_obj = program_obj
    res._scalar_type = scalar_type
    res._rows_length = rows_length
    res._row_length = row_length
    res._normalizable = normalizable
    res._location = location
    res._array_length = array_length
    res._dimension = dimension
    res._shape = shape
    res._name = name
    return res


def make_uniform(name, gl_type, program_obj, location, array_length, ctx):
    tmp = UNIFORM_LOOKUP_TABLE.get(gl_type, (False, 1, 4, '1i'))
    matrix, dimension, element_size, fmt = tmp
    res = Uniform()
    res._name = name
    res._type = gl_type
    res._fmt = fmt
    res._program_obj = program_obj
    res._location = location
    res._array_length = array_length
    res._matrix = matrix
    res._dimension = dimension
    res._element_size = element_size
    res._ctx = ctx
    return res


def make_uniform_block(name, program_obj, index, size, ctx):
    res = UniformBlock()
    res._name = name
    res._program_obj = program_obj
    res._index = index
    res._size = size
    res._ctx = ctx
    return res


def make_subroutine(name, index):
    res = Subroutine()
    res._name = name
    res._index = index
    return res


def make_varying(name, number, array_length, dimension):
    res = Varying()
    res._name = name
    res._array_length = array_length
    res._dimension = dimension
    return res


class InvalidObject:
    pass



from unittest import TestCase

import pytest
import moderngl
import numpy
import platform


class ContextTests(TestCase):

    def test_create_destroy(self):
        """Create and destroy a context"""
        for _ in range(25):
            ctx = moderngl.create_context(standalone=True)
            ctx.release()

    def test_context_switch(self):
        """Ensure context switching is working"""
        ctx1 = moderngl.create_context(standalone=True)
        ctx2 = moderngl.create_context(standalone=True)

        with ctx1 as ctx:
            buffer1 = ctx.buffer(reserve=1024)
        
        with ctx2 as ctx:
            buffer2 = ctx.buffer(reserve=1024)

        self.assertEqual(buffer1.glo, buffer2.glo)
        ctx1.release()
        ctx2.release()

    def test_exit(self):
        """Ensure the previous context was activated on exit"""
        ctx1 = moderngl.create_context(standalone=True)
        ctx2 = moderngl.create_context(standalone=True)

        with ctx1 as ctx:
            ctx.buffer(reserve=1024)

        # Will error out if no context is active "moderngl.error.Error: cannot create buffer"
        ctx1.buffer(reserve=1024)

        ctx1.release()
        ctx2.release()

    def test_share(self):
        """Create resources with shared context"""
        if platform.system().lower() in ["darwin", "linux"]:
            self.skipTest('Context sharing not supported on darwin')

        data1 = numpy.array([1, 2, 3, 4], dtype='u1')
        data2 = numpy.array([4, 3, 2, 1], dtype='u1')

        ctx1 = moderngl.create_context(standalone=True)
        ctx2 = moderngl.create_context(standalone=True, share=True)

        with ctx1 as ctx:
            b1 = ctx.buffer(data=data1)

        with ctx2 as ctx:
            b2 = ctx.buffer(data=data2)

        # Because the resources are shared the name should increment
        self.assertEqual(b1.glo, 1)
        self.assertEqual(b2.glo, 2)

        # Ensure we can read the same buffer data in both contexts
        with ctx1:
            self.assertEqual(b1.read(), b'\x01\x02\x03\x04')
            self.assertEqual(b2.read(), b'\x04\x03\x02\x01')

        with ctx2:
            self.assertEqual(b1.read(), b'\x01\x02\x03\x04')
            self.assertEqual(b2.read(), b'\x04\x03\x02\x01')

        ctx1.release()
        ctx2.release()

    def test_extensions(self):
        ctx = moderngl.create_context(standalone=True)
        # self.assertTrue("GL_ARB_vertex_array_object" in ctx.extensions)
        # self.assertTrue("GL_ARB_transform_feedback2" in ctx.extensions)
        # self.assertTrue("GL_ARB_shader_subroutine" in ctx.extensions)
        self.assertIsInstance(ctx.extensions, set)
        self.assertTrue(len(ctx.extensions) > 0)
        ctx.release()

    def test_attributes(self):
        """Ensure enums are present in the context instance"""
        ctx = moderngl.create_context(standalone=True)
        # Flags
        self.assertIsInstance(ctx.NOTHING, int)
        self.assertIsInstance(ctx.BLEND, int)
        self.assertIsInstance(ctx.DEPTH_TEST, int)
        self.assertIsInstance(ctx.CULL_FACE, int)
        self.assertIsInstance(ctx.RASTERIZER_DISCARD, int)
        self.assertIsInstance(ctx.PROGRAM_POINT_SIZE, int)

        # Primitive modes
        self.assertIsInstance(ctx.POINTS, int)
        self.assertIsInstance(ctx.LINES, int)
        self.assertIsInstance(ctx.LINE_LOOP, int)
        self.assertIsInstance(ctx.LINE_STRIP, int)
        self.assertIsInstance(ctx.TRIANGLES, int)
        self.assertIsInstance(ctx.TRIANGLE_STRIP, int)
        self.assertIsInstance(ctx.TRIANGLE_FAN, int)
        self.assertIsInstance(ctx.LINES_ADJACENCY, int)
        self.assertIsInstance(ctx.LINE_STRIP_ADJACENCY, int)
        self.assertIsInstance(ctx.TRIANGLES_ADJACENCY, int)
        self.assertIsInstance(ctx.TRIANGLE_STRIP_ADJACENCY, int)
        self.assertIsInstance(ctx.PATCHES, int)

        # Texture filters
        self.assertIsInstance(ctx.LINEAR, int)
        self.assertIsInstance(ctx.NEAREST, int)
        self.assertIsInstance(ctx.NEAREST_MIPMAP_NEAREST, int)
        self.assertIsInstance(ctx.LINEAR_MIPMAP_LINEAR, int)
        self.assertIsInstance(ctx.LINEAR_MIPMAP_NEAREST, int)
        self.assertIsInstance(ctx.NEAREST_MIPMAP_LINEAR, int)

        # Blend functions
        self.assertIsInstance(ctx.ZERO, int)
        self.assertIsInstance(ctx.ONE, int)
        self.assertIsInstance(ctx.SRC_COLOR, int)
        self.assertIsInstance(ctx.ONE_MINUS_SRC_COLOR, int)
        self.assertIsInstance(ctx.SRC_ALPHA, int)
        self.assertIsInstance(ctx.ONE_MINUS_SRC_ALPHA, int)
        self.assertIsInstance(ctx.DST_ALPHA, int)
        self.assertIsInstance(ctx.ONE_MINUS_DST_ALPHA, int)
        self.assertIsInstance(ctx.DST_COLOR, int)
        self.assertIsInstance(ctx.ONE_MINUS_DST_COLOR, int)

        # Blend shortcuts
        self.assertIsInstance(ctx.DEFAULT_BLENDING, tuple)
        self.assertIsInstance(ctx.ADDITIVE_BLENDING, tuple)
        self.assertIsInstance(ctx.PREMULTIPLIED_ALPHA, tuple)

        # Blend equations
        self.assertIsInstance(ctx.FUNC_ADD, int)
        self.assertIsInstance(ctx.FUNC_SUBTRACT, int)
        self.assertIsInstance(ctx.FUNC_REVERSE_SUBTRACT, int)
        self.assertIsInstance(ctx.MIN, int)
        self.assertIsInstance(ctx.MAX, int)

        # Provoking vertex
        self.assertIsInstance(ctx.FIRST_VERTEX_CONVENTION, int)
        self.assertIsInstance(ctx.LAST_VERTEX_CONVENTION, int)

    def test_enable_direct(self):
        ctx = moderngl.create_context(standalone=True)
        ctx.error  # consume error during initialization
        # We already support this, but it's a safe value
        GL_PROGRAM_POINT_SIZE = 0x8642

        ctx.enable_direct(GL_PROGRAM_POINT_SIZE)
        self.assertEqual(ctx.error, "GL_NO_ERROR")

        ctx.disable_direct(GL_PROGRAM_POINT_SIZE)
        self.assertEqual(ctx.error, "GL_NO_ERROR")

    def test_info(self):
        ctx = moderngl.create_context(standalone=True)
        self.assertIsInstance(ctx.info, dict)
        self.assertTrue(len(ctx.info) > 50)
        self.assertTrue(ctx.info["GL_MAX_GEOMETRY_OUTPUT_VERTICES"] >= 256)

    def test_polyon_offset(self):
        ctx = moderngl.create_context(standalone=True)
        ctx.polygon_offset = 0.0, 1.0
        self.assertEqual(ctx.polygon_offset, (0.0, 1.0))
        ctx.polygon_offset = 1.0, 0.0
        self.assertEqual(ctx.polygon_offset, (1.0, 0.0))
        ctx.polygon_offset = -1.0, -1.0
        self.assertEqual(ctx.polygon_offset, (-1.0, -1.0))

    def test_context_gc_modes(self):
        """Check gc mode default and set different modes"""
        ctx = moderngl.create_context(standalone=True)
        # gc mode should be None by default
        self.assertIsNone(ctx.gc_mode)
        ctx.gc_mode = None
        ctx.gc_mode = "auto"
        ctx.gc_mode = "context_gc"
        with pytest.raises(ValueError):
            ctx.gc_mode = "something"
        ctx.release()

    def test_context_gc(self):
        """Simple usage of context_gc"""
        ctx = moderngl.create_context(standalone=True)
        ctx.gc_mode = "context_gc"

        # Buffer
        buff = ctx.buffer(reserve=1024)
        buff = None
        self.assertEqual(ctx.gc(), 1)

        # Texture
        tex = ctx.texture((10, 10), 4)
        tex = None
        self.assertEqual(ctx.gc(), 1)

        # Texture Arrray
        tex_array = ctx.texture_array((10, 10, 10), 4)
        tex_array = None
        self.assertEqual(ctx.gc(), 1)

        # Texture cube
        tex_cube = ctx.texture_cube((10, 10), 4)
        tex_cube = None
        self.assertEqual(ctx.gc(), 1)

        # Texture3D
        tex_3d = ctx.texture3d((10, 10, 10), 4)
        tex_3d = None
        self.assertEqual(ctx.gc(), 1)

        # Sampler
        samp = ctx.sampler()
        samp = None
        self.assertEqual(ctx.gc(), 1)

        # Renderbuffer
        rb = ctx.renderbuffer((10, 19))
        rb = None
        self.assertEqual(ctx.gc(), 1)

        # Program
        prog = ctx.program(
            vertex_shader="""
            #version 330
            out float value;
            void main() {
                value = float(gl_VertexID);
            }
            """
        )
        prog = None
        self.assertEqual(ctx.gc(), 1)

        # Framebuffer
        fbo = ctx.framebuffer(
            color_attachments=[ctx.texture((10, 10), 4)],
            depth_attachment=ctx.depth_texture((10, 10))
        )
        fbo = None
        self.assertEqual(ctx.gc(), 3)

        # # Compute Shader
        # cs = ctx.compute_shader(
        #     """
        #     #version 450
            
        #     layout(local_size_x=4, local_size_y=4) in;

        #     layout(rgba8, binding=0) uniform imageCube img_in;
        #     layout(rgba8, binding=1) uniform imageCube img_out;

        #     void main() {
        #         for (int i = 0; i < 6; i++) {
        #             vec4 fragment = imageLoad(img_in, ivec3(gl_LocalInvocationID.xy, i));
        #             imageStore(img_out, ivec3(gl_LocalInvocationID.xy, i), fragment);
        #         }
        #     }
        #     """
        # )
        # cs = None
        # self.assertEqual(ctx.gc(), 1)

        ctx.release()



'''
example of using compute shader.

requirements:
    - numpy
    - pillow (for output)
'''
from string import Template
import moderngl
import numpy as np
import PIL.Image


COMPUTE_SOURCE = """
// author: minu jeong

#version 430

#define X $X
#define Y $Y
#define Z $Z
#define W $W
#define H $H

layout(local_size_x=X, local_size_y=Y, local_size_z=Z) in;
layout (std430, binding=0) buffer in_0
{
    vec4 inxs[1];
};

layout (std430, binding=1) buffer out_0
{
    vec4 outxs[1];
};

layout (std430, binding=2) buffer uv_0
{
    vec2 uvs[1];
};

#define win_width 5
#define win_height 5
#define win_wh 25
vec4 window[win_wh] = {
    // should manually initialize this
    vec4(0), vec4(0), vec4(0), vec4(0), vec4(0),
    vec4(0), vec4(0), vec4(0), vec4(0), vec4(0),
    vec4(0), vec4(0), vec4(0), vec4(0), vec4(0),
    vec4(0), vec4(0), vec4(0), vec4(0), vec4(0),
    vec4(0), vec4(0), vec4(0), vec4(0), vec4(0)
};

void main()
{
    // define consts
    const int x = int(gl_LocalInvocationID.x);
    const int y = int(gl_WorkGroupID.x);
    const int frag_i = x + y * W;

    int ignored = 0;
    // read window
    for (int win_x = 0; win_x < win_width; win_x++)
    {
        for (int win_y = 0; win_y < win_height; win_y++)
        {
            int win_i = win_y * win_width + win_x;
            int wox = win_x - win_width / 2;
            int woy = win_y - win_height / 2;
            int src_i = x + wox + (y + woy) * W;
            if (src_i < 0 || src_i > W * H)
            {
                window[win_i] = vec4(0, 0, 0, 0);
                ignored++;
                continue;
            }

            window[win_i] = inxs[src_i];
        }
    }

    // simple bubble sort to find median
    while(true)
    {
        bool is_swapped = false;
        for (int win_ii = win_wh - 1; win_ii > 1; win_ii--)
        {
            vec4 now = window[win_ii];
            if (now.w == 0.0) { continue; }
            if (length(window[win_ii - 1]) > length(now))
            {
                // swap
                window[win_ii] = window[win_ii - 1];
                window[win_ii - 1] = now;
                is_swapped = true;
            }
        }

        if (!is_swapped)
        {
            break;
        }
    }
    int median_i = win_wh / 2 + ignored / 2;
    vec4 median = window[median_i];

    // write to buffer
    outxs[frag_i] = vec4(median.xyz, 1.0);
}
"""

# W = X * Y  // for each run, handles a row of pixels
# execute compute shader for H times to complete
W = 512
H = 256
X = W
Y = 1
Z = 1
FRAMES = 50

context = moderngl.create_standalone_context(require=430)
compute_shader = context.compute_shader(
    Template(COMPUTE_SOURCE).substitute(
        W=W,
        H=H,
        X=X + 1,
        Y=Y,
        Z=Z,
    )
)

# init buffers
buffer_a = context.buffer(data=np.random.uniform(0.0, 1.0, (H, W, 4)).astype("f4"))
buffer_b = context.buffer(data=np.zeros((H, W, 4)).astype('f4'))

imgs = []
last_buffer = buffer_b

for i in range(FRAMES):
    toggle = True if i % 2 else False
    buffer_a.bind_to_storage_buffer(1 if toggle else 0)
    buffer_b.bind_to_storage_buffer(0 if toggle else 1)

    # toggle 2 buffers as input and output
    last_buffer = buffer_a if toggle else buffer_b

    # local invocation id x -> pixel x
    # work group_id x -> pixel y
    # eg) buffer[x, y] = gl_LocalInvocationID.x + gl_WorkGroupID.x * W
    compute_shader.run(group_x=H, group_y=1)

    # print out
    output = np.frombuffer(last_buffer.read(), dtype=np.float32)
    # output = output.reshape((H, W, 4))
    output = np.multiply(output, 255).astype(np.uint8)
    imgs.append(PIL.Image.frombuffer("RGBA", (W, H), output))


out_path = "debug.gif"
print("saving to", out_path)
imgs[0].save(out_path, save_all=True, append_images=imgs[1:], duration=15, loop=0)



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import math
# import random
# import struct

# import GLWindow
# import ModernGL

# # Window & Context

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         uniform vec2 Screen;
#         in vec2 vert;

#         void main() {
#             gl_Position = vec4((vert / Screen) * 2.0 - 1.0, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         out vec4 color;

#         void main() {
#             color = vec4(0.30, 0.50, 1.00, 1.0);
#         }
#     ''')
# ])

# tvert = ctx.vertex_shader('''
#     #version 330

#     in vec2 in_pos;
#     in vec2 in_prev;

#     out vec2 out_pos;
#     out vec2 out_prev;

#     void main() {
#         out_pos = in_pos * 2.0 - in_prev;
#         out_prev = in_pos;
#     }
# ''')

# transform = ctx.program(tvert, ['out_pos', 'out_prev'])


# def particle():
#     a = random.uniform(0.0, math.pi * 2.0)
#     r = random.uniform(0.0, 1.0)

#     cx, cy = wnd.mouse[0], wnd.mouse[1]

#     return struct.pack('2f2f', cx, cy, cx + math.cos(a) * r, cy + math.sin(a) * r)


# vbo1 = ctx.buffer(b''.join(particle() for i in range(1024)))
# vbo2 = ctx.buffer(reserve=vbo1.size)

# vao1 = ctx.simple_vertex_array(transform, vbo1, ['in_pos', 'in_prev'])
# vao2 = ctx.simple_vertex_array(transform, vbo2, ['in_pos', 'in_prev'])

# render_vao = ctx.vertex_array(prog, [
#     (vbo1, '2f8x', ['vert']),
# ])

# idx = 0

# ctx.point_size = 5.0

# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)

#     prog.uniforms['Screen'].value = wnd.size

#     for i in range(8):
#         vbo1.write(particle(), offset=idx * struct.calcsize('2f2f'))
#         idx = (idx + 1) % 1024

#     render_vao.render(ModernGL.POINTS, 1024)
#     vao1.transform(vbo2, ModernGL.POINTS, 1024)
#     ctx.copy_buffer(vbo1, vbo2)



import struct

import ModernGL
from kivy.app import App
from kivy.graphics import Callback
from kivy.uix.widget import Widget


class CustomWidget(Widget):
    def __init__(self, **kwargs):
        super(CustomWidget, self).__init__(**kwargs)

        with self.canvas:

            self.ctx = ModernGL.create_context()

            self.vert = vertex_shader='''
                #version 330

                in vec2 vert;

                void main() {
                    gl_Position = vec4(vert, 0.0, 1.0);
                }
            ''')

            self.frag = fragment_shader='''
                #version 330

                out vec4 color;

                void main() {
                    color = vec4(0.30, 0.50, 1.00, 1.0);
                }
            ''')

            self.prog = self.ctx.program(self.vert, self.frag])

            self.vbo = self.ctx.buffer(struct.pack('6f', 0.0, 0.8, -0.6, -0.8, 0.6, -0.8))
            self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, ['vert'])

            self.draw()

            Callback(self.draw)

    def draw(self, *args):
        self.ctx.clear(0.9, 0.9, 0.9)
        self.vao.render()

    def ask_update(self, *args):
        self.canvas.ask_update()


class MainApp(App):
    def build(self):
        return CustomWidget()


if __name__ == '__main__':
    MainApp().run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct
# import random

# import GLWindow
# import ModernGL

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 vert;
#         in vec3 tex_coord;
#         out vec3 v_tex_coord;

#         void main() {
#             gl_Position = vec4(vert, 0.0, 1.0);
#             v_tex_coord = tex_coord;
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         uniform samplerCube Texture;

#         in vec3 v_tex_coord;
#         out vec4 color;

#         void main() {
#             color = vec4(texture(Texture, v_tex_coord).rgb, 1.0);
#         }
#     '''),
# ])

# data = bytes()

# for i in range(11):
#     data += struct.pack('5f', i / 10.0 - 0.5, 0.2, random.uniform(-1, 1), random.uniform(-1, 1), random.uniform(-1, 1))
#     data += struct.pack('5f', i / 10.0 - 0.5, -0.2, random.uniform(-1, 1), random.uniform(-1, 1), random.uniform(-1, 1))

# vbo = ctx.buffer(data)

# vao = ctx.simple_vertex_array(prog, vbo, ['vert', 'tex_coord'])

# pixels = bytes()

# for i in range(4):
#     for j in range(4):
#         for k in range(6):
#             r = int(i * 255 / 3)
#             g = int(j * 255 / 3)
#             b = int(k * 255 / 5)
#             pixels += struct.pack('3B', r, g, b)

# texture = ctx.texture_cube((4, 4), 3, pixels)
# texture.use()

# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)
#     vao.render(ModernGL.TRIANGLE_STRIP)



from typing import Any, Optional, Tuple, Union

from _moderngl import InvalidObject

from .buffer import Buffer

__all__ = ['Texture3D']


class Texture3D:
    """
    A Texture is an OpenGL object that contains one or more images that all have the same image format.

    A texture can be used in two ways. It can
    be the source of a texture access from a Shader, or it can be used
    as a render target.

    A Texture3D object cannot be instantiated directly, it requires a context.
    Use :py:meth:`Context.texture3d` to create one.
    """

    __slots__ = ['mglo', '_size', '_components', '_samples', '_dtype', '_glo', 'ctx', 'extra']

    def __init__(self):
        self.mglo = None  #: Internal representation for debug purposes only.
        self._size = (None, None, None)
        self._components = None
        self._samples = None
        self._dtype = None
        self._glo = None
        self.ctx = None  #: The context this object belongs to
        self.extra = None  #: Any - Attribute for storing user defined objects
        raise TypeError()

    def __repr__(self):
        if hasattr(self, '_glo'):
            return f"<{self.__class__.__name__}: {self._glo}>"
        else:
            return f"<{self.__class__.__name__}: INCOMPLETE>"

    def __eq__(self, other: Any):
        return type(self) is type(other) and self.mglo is other.mglo

    def __hash__(self) -> int:
        return id(self)

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def repeat_x(self) -> bool:
        """
        bool: The x repeat flag for the texture (Default ``True``).

        Example::

            # Enable texture repeat (GL_REPEAT)
            texture.repeat_x = True

            # Disable texture repeat (GL_CLAMP_TO_EDGE)
            texture.repeat_x = False
        """
        return self.mglo.repeat_x

    @repeat_x.setter
    def repeat_x(self, value: bool) -> None:
        self.mglo.repeat_x = value

    @property
    def repeat_y(self) -> bool:
        """
        bool: The y repeat flag for the texture (Default ``True``).

        Example::

            # Enable texture repeat (GL_REPEAT)
            texture.repeat_y = True

            # Disable texture repeat (GL_CLAMP_TO_EDGE)
            texture.repeat_y = False
        """
        return self.mglo.repeat_y

    @repeat_y.setter
    def repeat_y(self, value: bool) -> None:
        self.mglo.repeat_y = value

    @property
    def repeat_z(self) -> bool:
        """
        bool: The z repeat flag for the texture (Default ``True``).

        Example::

            # Enable texture repeat (GL_REPEAT)
            texture.repeat_z = True

            # Disable texture repeat (GL_CLAMP_TO_EDGE)
            texture.repeat_z = False
        """
        return self.mglo.repeat_z

    @repeat_z.setter
    def repeat_z(self, value: bool) -> None:
        self.mglo.repeat_z = value

    @property
    def filter(self) -> Tuple[int, int]:
        """tuple: The filter of the texture."""
        return self.mglo.filter

    @filter.setter
    def filter(self, value: Tuple[int, int]) -> None:
        self.mglo.filter = value

    @property
    def swizzle(self) -> str:
        """
        str: The swizzle mask of the texture (Default ``'RGBA'``).

        The swizzle mask change/reorder the ``vec4`` value returned by the ``texture()`` function
        in a GLSL shaders. This is represented by a 4 character string were each
        character can be::

            'R' GL_RED
            'G' GL_GREEN
            'B' GL_BLUE
            'A' GL_ALPHA
            '0' GL_ZERO
            '1' GL_ONE

        Example::

            # Alpha channel will always return 1.0
            texture.swizzle = 'RGB1'

            # Only return the red component. The rest is masked to 0.0
            texture.swizzle = 'R000'

            # Reverse the components
            texture.swizzle = 'ABGR'
        """
        return self.mglo.swizzle

    @swizzle.setter
    def swizzle(self, value: str) -> None:
        self.mglo.swizzle = value

    @property
    def width(self) -> int:
        """int: The width of the texture."""
        return self._size[0]

    @property
    def height(self) -> int:
        """int: The height of the texture."""
        return self._size[1]

    @property
    def depth(self) -> int:
        """int: The depth of the texture."""
        return self._size[2]

    @property
    def size(self) -> tuple:
        """tuple: The size of the texture."""
        return self._size

    @property
    def components(self) -> int:
        """int: The number of components of the texture."""
        return self._components

    @property
    def dtype(self) -> str:
        """str: Data type."""
        return self._dtype

    @property
    def glo(self) -> int:
        """
        int: The internal OpenGL object.

        This values is provided for debug purposes only.
        """
        return self._glo

    def read(self, *, alignment: int = 1) -> bytes:
        """
        Read the pixel data as bytes into system memory.

        Keyword Args:
            alignment (int): The byte alignment of the pixels.

        Returns:
            bytes
        """
        return self.mglo.read(alignment)

    def read_into(
        self,
        buffer: Any,
        *,
        alignment: int = 1,
        write_offset: int = 0,
    ) -> None:
        """
        Read the content of the texture into a bytearray or :py:class:`~moderngl.Buffer`.

        The advantage of reading into a :py:class:`~moderngl.Buffer` is that pixel data
        does not need to travel all the way to system memory::

            # Reading pixel data into a bytearray
            data = bytearray(8)
            texture = ctx.texture3d((2, 2, 2), 1)
            texture.read_into(data)

            # Reading pixel data into a buffer
            data = ctx.buffer(reserve=8)
            texture = ctx.texture3d((2, 2, 2), 1)
            texture.read_into(data)

        Args:
            buffer (Union[bytearray, Buffer]): The buffer that will receive the pixels.

        Keyword Args:
            alignment (int): The byte alignment of the pixels.
            write_offset (int): The write offset.
        """
        if type(buffer) is Buffer:
            buffer = buffer.mglo

        return self.mglo.read_into(buffer, alignment, write_offset)

    def write(
        self,
        data: Any,
        viewport: Optional[Union[Tuple[int, int, int], Tuple[int, int, int, int, int, int]]] = None,
        *,
        alignment: int = 1,
    ) -> None:
        r"""
        Update the content of the texture from byte data or a moderngl :py:class:`~moderngl.Buffer`.

        Examples::

            # Write data from a moderngl Buffer
            data = ctx.buffer(reserve=8)
            texture = ctx.texture3d((2, 2, 2), 1)
            texture.write(data)

            # Write data from bytes
            data = b'\xff\xff\xff\xff\xff\xff\xff\xff'
            texture = ctx.texture3d((2, 2), 1)
            texture.write(data)

        Args:
            data (bytes): The pixel data.
            viewport (tuple): The viewport.

        Keyword Args:
            alignment (int): The byte alignment of the pixels.
        """
        if type(data) is Buffer:
            data = data.mglo

        self.mglo.write(data, viewport, alignment)

    def build_mipmaps(self, base: int = 0, max_level: int = 1000) -> None:
        """
        Generate mipmaps.

        This also changes the texture filter to ``LINEAR_MIPMAP_LINEAR, LINEAR``
        (Will be removed in ``6.x``)

        Keyword Args:
            base (int): The base level
            max_level (int): The maximum levels to generate
        """
        self.mglo.build_mipmaps(base, max_level)

    def use(self, location: int = 0) -> None:
        """
        Bind the texture to a texture unit.

        The location is the texture unit we want to bind the texture.
        This should correspond with the value of the ``sampler3D``
        uniform in the shader because samplers read from the texture
        unit we assign to them::

            # Define what texture unit our two sampler3D uniforms should represent
            program['texture_a'] = 0
            program['texture_b'] = 1
            # Bind textures to the texture units
            first_texture.use(location=0)
            second_texture.use(location=1)

        Args:
            location (int): The texture location/unit.
        """
        self.mglo.use(location)

    def bind_to_image(self, unit: int, read: bool = True, write: bool = True, level: int = 0, format: int = 0) -> None:
        """
        Bind a texture to an image unit (OpenGL 4.2 required).

        This is used to bind textures to image units for shaders.
        The idea with image load/store is that the user can bind
        one of the images in a Texture to a number of image binding points
        (which are separate from texture image units). Shaders can read
        information from these images and write information to them,
        in ways that they cannot with textures.

        It's important to specify the right access type for the image.
        This can be set with the ``read`` and ``write`` arguments.
        Allowed combinations are:

        - **Read-only**: ``read=True`` and ``write=False``
        - **Write-only**: ``read=False`` and ``write=True``
        - **Read-write**: ``read=True`` and ``write=True``

        ``format`` specifies the format that is to be used when performing
        formatted stores into the image from shaders. ``format`` must be
        compatible with the texture's internal format. **By default the format
        of the texture is passed in. The format parameter is only needed
        when overriding this behavior.**

        Note that we bind the 3D textured layered making the entire texture
        readable and writable. It is possible to bind a specific 2D section
        in the future.

        More information:

        - https://www.khronos.org/opengl/wiki/Image_Load_Store
        - https://www.khronos.org/registry/OpenGL-Refpages/gl4/html/glBindImageTexture.xhtml

        Args:
            unit (int): Specifies the index of the image unit to which to bind the texture
            texture (:py:class:`moderngl.Texture`): The texture to bind
        Keyword Args:
            read (bool): Allows the shader to read the image (default: ``True``)
            write (bool): Allows the shader to write to the image (default: ``True``)
            level (int): Level of the texture to bind (default: ``0``).
            format (int): (optional) The OpenGL enum value representing the format (defaults to the texture's format)
        """
        self.mglo.bind(unit, read, write, level, format)

    def release(self) -> None:
        """Release the ModernGL object."""
        if self.mglo is not None:
            self.mglo.release()
            self.mglo = InvalidObject()



import moderngl
import numpy as np
from pyrr import Matrix44

from ported._example import Example


class CrateExample(Example):
    title = "Crate"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 Mvp;

                in vec3 in_position;
                in vec3 in_normal;
                in vec2 in_texcoord_0;

                out vec3 v_vert;
                out vec3 v_norm;
                out vec2 v_text;

                void main() {
                    gl_Position = Mvp * vec4(in_position, 1.0);
                    v_vert = in_position;
                    v_norm = in_normal;
                    v_text = in_texcoord_0;
                }
            ''',
            fragment_shader='''
                #version 330

                uniform vec3 Light;
                uniform sampler2D Texture;

                in vec3 v_vert;
                in vec3 v_norm;
                in vec2 v_text;

                out vec4 f_color_a;
                out vec4 f_color_b;
                out vec4 f_color_c;

                void main() {
                    float lum = clamp(dot(normalize(Light - v_vert), normalize(v_norm)), 0.0, 1.0) * 0.8 + 0.2;
                    f_color_a = vec4(texture(Texture, v_text).rgb * lum, 1.0);
                    f_color_b = vec4(texture(Texture, v_text).rgb * lum * vec3(0.5, 1.1, 0.6), 1.0);
                    f_color_c = vec4(texture(Texture, v_text).rgb * lum * vec3(0.5, 0.6, 1.1), 1.0);
                }
            ''',
            fragment_outputs={
                'f_color_a': 2,
                'f_color_b': 1,
                'f_color_c': 0,
            },
        )

        self.mvp = self.prog['Mvp']
        self.light = self.prog['Light']

        self.scene = self.load_scene('crate.obj')
        self.vao = self.scene.root_nodes[0].mesh.vao.instance(self.prog)
        self.texture = self.load_texture_2d('crate.png')

    def render(self, time, frame_time):
        angle = time
        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.DEPTH_TEST)

        camera_pos = (np.cos(angle) * 3.0, np.sin(angle) * 3.0, 2.0)

        proj = Matrix44.perspective_projection(45.0, self.aspect_ratio, 0.1, 100.0)
        lookat = Matrix44.look_at(
            camera_pos,
            (0.0, 0.0, 0.5),
            (0.0, 0.0, 1.0),
        )

        self.mvp.write((proj * lookat).astype('f4'))
        self.light.value = camera_pos
        self.texture.use()
        self.vao.render()


if __name__ == '__main__':
    CrateExample.run()



'''
    Renders a blue triangle
'''

import numpy as np

from ported._example import Example


class HelloWorld(Example):
    gl_version = (3, 3)
    title = "Hello World"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 f_color;

                void main() {
                    f_color = vec4(0.3, 0.5, 1.0, 1.0);
                }
            ''',
        )

        vertices = np.array([
            0.0, 0.8,
            -0.6, -0.8,
            0.6, -0.8,
        ], dtype='f4')

        self.vbo = self.ctx.buffer(vertices)
        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, 'in_vert')

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.vao.render()


if __name__ == '__main__':
    HelloWorld.run()



import struct

import ModernGL
from kivy.app import App
from kivy.core.window import Window
from kivy.graphics import Callback
from kivy.uix.widget import Widget


class CustomWidget(Widget):
    def __init__(self, **kwargs):
        super(CustomWidget, self).__init__(**kwargs)

        with self.canvas:
            self.ctx = ModernGL.create_context()

            self.prog = self.ctx.program(
                vertex_shader='''
                    #version 330

                    uniform vec2 WindowSize;

                    in vec2 in_vert;
                    in vec3 in_color;

                    out vec3 v_color;

                    void main() {
                        v_color = in_color;
                        gl_Position = vec4(in_vert / WindowSize * 2.0, 0.0, 1.0);
                    }
                '''),
                fragment_shader='''
                    #version 330

                    in vec3 v_color;
                    out vec4 f_color;

                    void main() {
                        f_color = vec4(v_color, 1.0);
                    }
                '''),
            ])

            self.window_size = self.prog.uniforms['WindowSize']

            self.vbo = self.ctx.buffer(struct.pack(
                '15f',
                0.0, 100.0, 1.0, 0.0, 0.0,
                -86.0, -50.0, 0.0, 1.0, 0.0,
                86.0, -50.0, 0.0, 0.0, 1.0,
            ))

            self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, ['in_vert', 'in_color'])

            Callback(self.draw)

    def draw(self, *args):
        self.width, self.height = Window.size
        self.ctx.viewport = (0, 0, self.width, self.height)
        self.ctx.clear(0.9, 0.9, 0.9)
        self.ctx.enable(ModernGL.BLEND)
        self.window_size.value = (self.width, self.height)
        self.vao.render()

    def ask_update(self, *args):
        self.canvas.ask_update()


class MainApp(App):
    def build(self):
        return CustomWidget()


if __name__ == '__main__':
    MainApp().run()



import unittest

from moderngl.mgl import fmtdebug


GL_BYTE = 0x1400
GL_UNSIGNED_BYTE = 0x1401
GL_SHORT = 0x1402
GL_UNSIGNED_SHORT = 0x1403
GL_INT = 0x1404
GL_UNSIGNED_INT = 0x1405
GL_FLOAT = 0x1406
GL_DOUBLE = 0x140A
GL_HALF_FLOAT = 0x140B


class TestBuffer(unittest.TestCase):

    def check(self, fmt, res):
        self.assertEqual(fmtdebug(fmt), res)

    def test_format_1(self):
        self.check('1f', (4, 1, 0, True, ((4, 1, GL_FLOAT, False),)))
        self.check('2f', (8, 1, 0, True, ((8, 2, GL_FLOAT, False),)))
        self.check('3f', (12, 1, 0, True, ((12, 3, GL_FLOAT, False),)))
        self.check('4f', (16, 1, 0, True, ((16, 4, GL_FLOAT, False),)))
        self.check('1i', (4, 1, 0, True, ((4, 1, GL_INT, False),)))
        self.check('2i', (8, 1, 0, True, ((8, 2, GL_INT, False),)))
        self.check('3i', (12, 1, 0, True, ((12, 3, GL_INT, False),)))
        self.check('4i', (16, 1, 0, True, ((16, 4, GL_INT, False),)))
        self.check('1u', (4, 1, 0, True, ((4, 1, GL_UNSIGNED_INT, False),)))
        self.check('2u', (8, 1, 0, True, ((8, 2, GL_UNSIGNED_INT, False),)))
        self.check('3u', (12, 1, 0, True, ((12, 3, GL_UNSIGNED_INT, False),)))
        self.check('4u', (16, 1, 0, True, ((16, 4, GL_UNSIGNED_INT, False),)))

    def test_format_2(self):
        self.check('1f1', (1, 1, 0, True, ((1, 1, GL_UNSIGNED_BYTE, True),)))
        self.check('1f2', (2, 1, 0, True, ((2, 1, GL_HALF_FLOAT, False),)))
        self.check('1f4', (4, 1, 0, True, ((4, 1, GL_FLOAT, False),)))
        self.check('1f8', (8, 1, 0, True, ((8, 1, GL_DOUBLE, False),)))

    def test_format_3(self):
        self.check('1f1f', (0, 0, 0, False, ()))
        self.check('3f3f4i', (0, 0, 0, False, ()))
        self.check(
            '3f 3f 4i',
            (40, 3, 0, True, (
                (12, 3, GL_FLOAT, False),
                (12, 3, GL_FLOAT, False),
                (16, 4, GL_INT, False)
            ))
        )

    def test_format_4(self):
        self.check('3f 3f /v', (24, 2, 0, True, ((12, 3, GL_FLOAT, False), (12, 3, GL_FLOAT, False))))
        self.check(
            '3f 12x 3f /v',
            (36, 2, 0, True, (
                (12, 3, GL_FLOAT, False),
                (12, 12, 0, False),
                (12, 3, GL_FLOAT, False)
            ))
        )

    def test_format_5(self):
        # vert, norm, text, color
        self.check(
            '3f 3f 2f 4f1',
            (36, 4, 0, True, (
                (12, 3, GL_FLOAT, False),
                (12, 3, GL_FLOAT, False),
                (8, 2, GL_FLOAT, False),
                (4, 4, GL_UNSIGNED_BYTE, True),
            ))
        )

        # vert, norm, text, color
        self.check(
            '3f4 3f4 2f4 4f1',
            (36, 4, 0, True, (
                (12, 3, GL_FLOAT, False),
                (12, 3, GL_FLOAT, False),
                (8, 2, GL_FLOAT, False),
                (4, 4, GL_UNSIGNED_BYTE, True),
            ))
        )

        # vert, norm, padding, color
        self.check(
            '3f 3f 2x4 4f1',
            (36, 3, 0, True, (
                (12, 3, GL_FLOAT, False),
                (12, 3, GL_FLOAT, False),
                (8, 2, 0, False),
                (4, 4, GL_UNSIGNED_BYTE, True),
            ))
        )

    def test_format_6(self):
        self.check('2f/i', (8, 1, 1, True, ((8, 2, GL_FLOAT, False),)))
        self.check('2f /i', (8, 1, 1, True, ((8, 2, GL_FLOAT, False),)))

    def test_format_7(self):
        self.check('2f 2x4/i', (16, 1, 1, True, ((8, 2, GL_FLOAT, False), (8, 2, 0, False))))
        self.check('2f 2x4 /i', (16, 1, 1, True, ((8, 2, GL_FLOAT, False), (8, 2, 0, False))))


if __name__ == '__main__':
    unittest.main()



import math
import random

from ported._example import Example


class EmptyWindow(Example):
    gl_version = (3, 3)
    title = "Empty Window"
    window_size = (1280, 720)
    aspect_ratio = 16 / 9
    resizable = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def render(self, time: float, frame_time: float):
        self.ctx.clear(
            (math.sin(time) + 1.0) / 2,
            (math.sin(time + 2) + 1.0) / 2,
            (math.sin(time + 3) + 1.0) / 2,
        )

    def resize(self, width: int, heigh: int):
        """
        Pick window resizes in case we need yo update
        internal states when this happens.
        """
        print("Window resized to", width, heigh)

    def iconify(self, iconify: bool):
        """Window hide/minimize and restore"""
        print("Window was iconified:", iconify)

    def key_event(self, key, action, modifiers):
        keys = self.wnd.keys

        # Key presses
        if action == keys.ACTION_PRESS:
            if key == keys.SPACE:
                print("SPACE key was pressed")

            # Using modifiers (shift and ctrl)

            if key == keys.Z and modifiers.shift:
                print("Shift + Z was pressed")

            if key == keys.Z and modifiers.ctrl:
                print("ctrl + Z was pressed")

        # Key releases
        elif action == self.wnd.keys.ACTION_RELEASE:
            if key == keys.SPACE:
                print("SPACE key was released")

        if action == keys.ACTION_PRESS:
            # Move the window around with AWSD
            if key == keys.A:
                self.wnd.position = self.wnd.position[0] - 20, self.wnd.position[1]
            if key == keys.D:
                self.wnd.position = self.wnd.position[0] + 20, self.wnd.position[1]
            if key == keys.W:
                self.wnd.position = self.wnd.position[0], self.wnd.position[1] - 20
            if key == keys.S:
                self.wnd.position = self.wnd.position[0], self.wnd.position[1] + 20

            # Resize window around with Shift + AWSD
            if self.wnd.modifiers.shift and key == keys.A:
                self.wnd.size = self.wnd.size[0] - 50, self.wnd.size[1]
            if self.wnd.modifiers.shift and key == keys.D:
                self.wnd.size = self.wnd.size[0] + 50, self.wnd.size[1]
            if self.wnd.modifiers.shift and key == keys.W:
                self.wnd.size = self.wnd.size[0], self.wnd.size[1] - 50
            if self.wnd.modifiers.shift and key == keys.S:
                self.wnd.size = self.wnd.size[0], self.wnd.size[1] + 50

            # toggle cursor
            if key == keys.C:
                self.wnd.cursor = not self.wnd.cursor

            # Shuffle window title
            if key == keys.T:
                title = list(self.wnd.title)
                random.shuffle(title)
                self.wnd.title = ''.join(title)

            # Toggle mouse exclusivity
            if key == keys.M:
                self.wnd.mouse_exclusivity = not self.wnd.mouse_exclusivity

    def mouse_position_event(self, x, y, dx, dy):
        print("Mouse position pos={} {} delta={} {}".format(x, y, dx, dy))

    def mouse_drag_event(self, x, y, dx, dy):
        print("Mouse drag pos={} {} delta={} {}".format(x, y, dx, dy))

    def mouse_scroll_event(self, x_offset, y_offet):
        print("mouse_scroll_event", x_offset, y_offet)

    def mouse_press_event(self, x, y, button):
        print("Mouse button {} pressed at {}, {}".format(button, x, y))
        print("Mouse states:", self.wnd.mouse_states)

    def mouse_release_event(self, x: int, y: int, button: int):
        print("Mouse button {} released at {}, {}".format(button, x, y))
        print("Mouse states:", self.wnd.mouse_states)

    def unicode_char_entered(self, char):
        print("unicode_char_entered:", char)


if __name__ == '__main__':
    EmptyWindow.run()



'''
    Renders a triangle that has all RGB combinations
'''

import numpy as np

from ported._example import Example


class SimpleColorTriangle(Example):
    gl_version = (3, 3)
    aspect_ratio = 16 / 9
    title = "Simple Color Triangle"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;

                in vec3 in_color;
                out vec3 v_color;    // Goes to the fragment shader

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                    v_color = in_color;
                }
            ''',
            fragment_shader='''
                #version 330

                in vec3 v_color;
                out vec4 f_color;

                void main() {
                    // We're not interested in changing the alpha value
                    f_color = vec4(v_color, 1.0);
                }
            ''',
        )

        # Point coordinates are put followed by the vec3 color values
        vertices = np.array([
            # x, y, red, green, blue
            0.0, 0.8, 1.0, 0.0, 0.0,
            -0.6, -0.8, 0.0, 1.0, 0.0,
            0.6, -0.8, 0.0, 0.0, 1.0,
        ], dtype='f4')

        self.vbo = self.ctx.buffer(vertices)

        # We control the 'in_vert' and `in_color' variables
        self.vao = self.ctx.vertex_array(
            self.prog,
            [
                # Map in_vert to the first 2 floats
                # Map in_color to the next 3 floats
                (self.vbo, '2f 3f', 'in_vert', 'in_color')
            ],
        )

    def render(self, time: float, frame_time: float):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.vao.render()


if __name__ == '__main__':
    SimpleColorTriangle.run()
